"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import logging
import warnings
from pathlib import Path

import geopandas as gpd
import networkx as nx
import osmnx
import pandas as pd
import pyproj

from ra2ce.network import networks_utils as nut
from ra2ce.network.exporters.network_exporter_factory import NetworkExporterFactory
from ra2ce.network.graph_files.graph_files_collection import GraphFilesCollection
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_wrappers.network_wrapper_factory import NetworkWrapperFactory


class Network:
    """Network in GeoDataFrame or NetworkX format.

    Networks can be created from shapefiles, OSM PBF files, can be downloaded from OSM online or can be loaded from
    feather or gpickle files. Origin-destination nodes can be added.

    Attributes:
        config: An object with the configuration details on how to create and adjust the network.
    """

    def __init__(
        self, network_config: NetworkConfigData, graph_files: GraphFilesCollection
    ):
        # General
        self._config_data = network_config
        self.project_name = network_config.project.name
        self.output_graph_dir = network_config.output_graph_dir
        if not self.output_graph_dir.is_dir():
            self.output_graph_dir.mkdir(parents=True)

        # Network
        self._network_dir = network_config.network_dir
        self.base_graph_crs = None  # Initiate variable
        self.base_network_crs = None  # Initiate variable

        self._network_config = network_config.network

        # Origins and destinations
        _origins_destinations = network_config.origins_destinations
        self.origins = _origins_destinations.origins
        self.destinations = _origins_destinations.destinations
        self.origins_names = _origins_destinations.origins_names
        self.destinations_names = _origins_destinations.destinations_names
        self.id_name_origin_destination = (
            _origins_destinations.id_name_origin_destination
        )
        self.origin_count = _origins_destinations.origin_count
        self.od_category = _origins_destinations.category
        self.region = _origins_destinations.region
        self.region_var = _origins_destinations.region_var

        # graph files
        self.graph_files = graph_files

    def add_od_nodes(self, graph: nx.Graph, crs: pyproj.CRS) -> nx.Graph:
        """Adds origins and destinations nodes from shapefiles to the graph.

        Args:
            graph (NetworkX graph): the NetworkX graph to which OD nodes should be added
            crs (int): the EPSG number of the coordinate reference system that is used

        Returns:
            graph (NetworkX graph): the NetworkX graph with OD nodes
        """
        from ra2ce.network.origins_destinations import (
            add_od_nodes,
            read_origin_destination_files,
        )

        name = "origin_destination_table"

        # Add the origin/destination nodes to the network
        ods = read_origin_destination_files(
            str(self.origins),
            self.origins_names,
            str(self.destinations),
            self.destinations_names,
            self.id_name_origin_destination,
            self.origin_count,
            crs,
            self.od_category,
            self.region if self.region else "",
            self.region_var,
        )

        (ods, graph) = add_od_nodes(ods, graph, crs, self.od_category)
        ods.crs = crs

        # Save the OD pairs (GeoDataFrame) as pickle
        ods.to_feather(self.output_graph_dir.joinpath(name + ".feather"), index=False)
        logging.info(f"Saved {name + '.feather'} in {self.output_graph_dir}.")

        # Save the OD pairs (GeoDataFrame) as shapefile
        if self._network_config.save_gpkg:
            ods_path = self.output_graph_dir.joinpath(name + ".gpkg")
            ods.to_file(ods_path, index=False)
            logging.info(f"Saved {ods_path.stem} in {ods_path.resolve().parent}.")

        return graph

    def generate_origins_from_raster(self):
        """Adds origins and destinations nodes from shapefiles to the graph."""
        from ra2ce.network.origins_destinations import origins_from_raster

        out_fn = origins_from_raster(
            self._network_dir,
            self._network_config.polygon,
            self.origins,
        )

        return out_fn

    def _export_network_files(
        self,
        network: nx.MultiGraph | gpd.GeoDataFrame,
        graph_type: str,
        types_to_export: list[str],
    ):
        _exporter = NetworkExporterFactory()
        _exporter.export(
            network=network,
            basename=graph_type,
            output_dir=self.output_graph_dir,
            export_types=types_to_export,
        )
        self.graph_files.set_file(_exporter.get_pickle_path())
        self.graph_files.set_graph(graph_type, network)

    def _include_attributes(self, attributes: list, graph: nx.Graph) -> nx.Graph:
        # If required_attributes are provided, check if all edges already have them
        if attributes and all(
            all(attr in edge_data for attr in attributes)
            for _, _, edge_data in graph.edges(data=True)
        ):
            # If all required attributes are present, return the original graph
            return graph
        graph = nut.add_x_y_to_nodes(graph)
        _, gdf_edges = osmnx.graph_to_gdfs(graph)
        updated_graph = copy.deepcopy(graph)
        for attribute in attributes:
            if attribute in gdf_edges.columns:
                attribute_values_gdf = gdf_edges[attribute]
                for (u, v, key), attribute_values in attribute_values_gdf.items():
                    updated_graph[u][v][key][attribute] = attribute_values
        return updated_graph

    def _get_new_network_and_graph(self, export_types: list[str]) -> None:
        """
        TODO: This method should be relying on a generic definition of a network result
        from `.get_network`. This means, instead of getting `_base_graph, _network_gdf`
        we get a generic `_ra2ce_network_wrapper` from which can later on just do a
        `.simplify_network` or `.add_eges`, etc. using inheritance.
        """

        _base_graph, _network_gdf = NetworkWrapperFactory(
            self._config_data
        ).get_network()

        self.base_graph_crs = _network_gdf.crs
        self.base_network_crs = _network_gdf.crs

        _base_graph = self._include_attributes(
            attributes=["avgspeed", "bridge", "tunnel"], graph=_base_graph
        )

        # Save the graph and geodataframe
        self._export_network_files(_base_graph, "base_graph", export_types)
        self._export_network_files(_network_gdf, "base_network", export_types)

    def _get_stored_network_and_graph(self) -> None:
        base_graph_filepath = self.graph_files.base_graph.file
        base_network_filepath = self.graph_files.base_network.file

        logging.info(
            "Apparently, you already did create a network with ra2ce earlier. "
            + "Ra2ce will use this: %s",
            base_graph_filepath,
        )

        def get_graph(
            file_type: str, file_path: Path | None
        ) -> nx.MultiGraph | gpd.GeoDataFrame:
            graph = self.graph_files.get_graph(file_type)
            if graph is None:
                raise FileNotFoundError(
                    "No base {} file found at {}.".format(file_type, file_path)
                )
            return graph

        _base_graph = get_graph("base_graph", base_graph_filepath)
        _network_gdf = get_graph("base_network", base_network_filepath)

        # Assuming the same CRS for both the network and graph
        self.base_graph_crs = _network_gdf.crs
        self.base_network_crs = _network_gdf.crs

    def create(self) -> GraphFilesCollection:
        """Handler function with the logic to call the right functions to create a network.

        Returns:
            (GraphFilesCollection): A collection of a network (GeoDataFrame) and 1 (base NetworkX graph) or 2 graphs (base NetworkX and OD graph)
        """
        # Save the 'base' network as gpickle and if the user requested, also as shapefile.
        to_save = (
            ["pickle"] if not self._network_config.save_gpkg else ["pickle", "gpkg"]
        )

        # For all graph and networks - check if it exists, otherwise, make the graph and/or network.
        if not (self.graph_files.base_graph.file or self.graph_files.base_network.file):
            self._get_new_network_and_graph(to_save)
        else:
            self._get_stored_network_and_graph()

        # create origins destinations graph
        if (
            (self.origins)
            and (self.destinations)
            and not self.graph_files.origins_destinations_graph.file
        ):
            # fetch the base graph
            base_graph = self.graph_files.base_graph.get_graph()
            # adding OD nodes
            if self.origins.suffix == ".tif":
                self.origins = self.generate_origins_from_raster()
            od_graph = self.add_od_nodes(base_graph, self.base_graph_crs)
            self._export_network_files(od_graph, "origins_destinations_graph", to_save)

        return self.graph_files



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import ast
import itertools
import logging
from abc import ABC, abstractmethod
from collections import defaultdict

import geopandas as gpd
import numpy as np
import pandas as pd

from ra2ce.analysis.losses.traffic_analysis.accumulated_traffic_dataclass import (
    AccumulatedTraffic,
)


class TrafficAnalysisBase(ABC):
    road_network: gpd.GeoDataFrame
    od_table: gpd.GeoDataFrame
    destinations_names: str

    def optimal_route_od_link(
        self,
    ) -> pd.DataFrame:
        """
        Gets the optimal routes based on utilitarian, egalitarian and prioritarian traffic.

        Returns:
            pd.DataFrame: Datafarme with the traffic indices for each of analysis.
        """
        unique_destination_nodes = np.unique(list(self.od_table["d_id"].fillna("0")))
        count_destination_nodes = len([x for x in unique_destination_nodes if x != "0"])

        _traffic = defaultdict(AccumulatedTraffic)
        for _, _road_network_row in self.road_network.iterrows():
            o_node, d_node = (
                _road_network_row["origin"],
                _road_network_row["destination"],
            )
            if "," in o_node:
                logging.error(
                    "List of nodes as 'origin node' is not accepted and will be skipped."
                )
                continue
            opt_path = self._get_opt_path_values(_road_network_row)
            for u_node, v_node in itertools.pairwise(opt_path):
                _nodes_key_name = self._get_node_key(u_node, v_node)
                _calculated_traffic = self._get_accumulated_traffic_from_node(
                    o_node, count_destination_nodes
                )

                if "," in d_node:
                    _calculated_traffic *= len(d_node.split(","))

                _traffic[_nodes_key_name] += _calculated_traffic

        return self._get_route_traffic(_traffic)

    def _get_node_key(self, u_node: str, v_node: str) -> str:
        return f"{u_node}_{v_node}"

    def _get_key_nodes(self, node_key: str) -> tuple[str, str]:
        return node_key.split("_")

    def _get_opt_path_values(self, _road_network_row: pd.Series) -> list[int]:
        _opt_path_value = _road_network_row["opt_path"]
        if isinstance(_opt_path_value, list):
            return _opt_path_value
        return ast.literal_eval(_opt_path_value)

    def _get_recorded_traffic_in_node(
        self, origin_node: str, column_name: str, count_destination_nodes: int
    ) -> float:
        return (
            self.od_table.loc[
                self.od_table["o_id"] == origin_node,
                column_name,
            ].values[0]
            / count_destination_nodes
        )

    @abstractmethod
    def _get_accumulated_traffic_from_node(
        self, target_node: str, total_d_nodes: int
    ) -> AccumulatedTraffic:
        raise NotImplementedError("Should be implemented in concrete class.")

    @abstractmethod
    def _get_route_traffic(
        self, traffic_data_wrapper: dict[str, AccumulatedTraffic]
    ) -> pd.DataFrame:
        raise NotImplementedError("Should be implemented in concrete class.")



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionDamages,
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.network.graph_files.graph_files_protocol import GraphFileProtocol
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.network_config_data.network_config_data import (
    OriginsDestinationsSection,
)


@dataclass
class AnalysisInputWrapper:
    analysis: AnalysisSectionDamages | AnalysisSectionLosses
    graph_file: Optional[GraphFileProtocol]
    graph_file_hazard: Optional[GraphFileProtocol]
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: Optional[HazardNames]
    origins_destinations: Optional[OriginsDestinationsSection]
    file_id: Optional[str]

    @classmethod
    def from_input(
        cls,
        analysis: AnalysisSectionDamages | AnalysisSectionLosses,
        analysis_config: AnalysisConfigWrapper,
        graph_file: Optional[GraphFileProtocol] = None,
        graph_file_hazard: Optional[GraphFileProtocol] = None,
    ) -> AnalysisInputWrapper:
        return cls(
            analysis=analysis,
            graph_file=graph_file,
            graph_file_hazard=graph_file_hazard,
            input_path=analysis_config.config_data.input_path,
            static_path=analysis_config.config_data.static_path,
            output_path=analysis_config.config_data.output_path,
            hazard_names=HazardNames.from_config(analysis_config),
            origins_destinations=analysis_config.config_data.origins_destinations,
            file_id=analysis_config.config_data.network.file_id,
        )



from __future__ import annotations

from enum import Enum


class Ra2ceEnumBase(Enum):
    """
    Base class for enums defined within Ra2ce.
    NONE = 0: Optional entry (config is optional and missing)
    INVALID = 99: Mandatory entry (config contains invalid value)
    """

    @classmethod
    def get_enum(cls, input: str | None) -> Ra2ceEnumBase:
        """
        Create an enum from a given input string.

        Args:
            input (str): Value from config.

        Returns:
            Ra2ceEnumBase: Enumeration instance.
            NONE: This entry is used if the config is missing.
            INVALID: This entry is used if the config value is invalid.
        """
        try:
            if not input:
                return cls.NONE
            return cls[input.upper().strip()]
        except (AttributeError, KeyError):
            return cls.INVALID

    def is_valid(self) -> bool:
        """
        Check if given value is valid.

        Args:
            key (str): Enum key (name)

        Returns:
            bool: If the given key is not a valid key
        """
        if self.name == "INVALID":
            return False
        return True

    @classmethod
    def list_valid_options(cls) -> list[Ra2ceEnumBase]:
        """
        List the enum options as allowed in the config.

        Returns:
            list[str | None]: Concatenated options, separated by ", "
        """
        return [_enum for _enum in cls][:-1]

    def __str__(self) -> str:
        """
        Overwrite the default __str__

        Returns:
            str: Value as in the config
        """
        return str(self.config_value)

    @property
    def config_value(self) -> str | None:
        """
        Reconstruct the name as it is known in the config.
        This could entail replacement of " " by "_" and lower() operations.

        Returns:
            str: Value as known in the config.
        """
        if self.name == "NONE":
            return None
        return self.name.lower()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import math
from pathlib import Path
from re import split
from statistics import mean

import networkx as nx

import ra2ce.network.networks_utils as nut
from ra2ce.network.avg_speed.avg_speed import AvgSpeed
from ra2ce.network.avg_speed.avg_speed_reader import AvgSpeedReader
from ra2ce.network.avg_speed.avg_speed_writer import AvgSpeedWriter
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


class AvgSpeedCalculator:
    """
    Class to calculate the average speed of a road network based on the maximum speed of the roads.
    """

    graph: nx.Graph
    avg_speed: AvgSpeed

    def __init__(self, graph: nx.Graph, output_graph_dir: Path | None) -> None:
        self.graph = graph
        self.avg_speed = self._calculate(output_graph_dir)

    @staticmethod
    def parse_speed(speed_input: float | str | list[str]) -> float:
        """
        Parse the average speed from the input string(s).

        Args:
            speed_input (float | str | list[str]): (List of) string(s) with the speed(s).
                Can have different formats, e.g. nan(float), 30.0, "50 mph", "50", "50;60", "50-60", "50|60".

        Returns:
            float: Average speed of the input string(s).
                0.0 is returned if the input is empty or can't be parsed correctly.
        """
        if not speed_input:
            return 0.0
        if isinstance(speed_input, float):
            if math.isnan(speed_input):
                return 0.0
            return speed_input
        if isinstance(speed_input, list):
            return mean(map(AvgSpeedCalculator.parse_speed, speed_input))
        if " mph" in speed_input:
            return float(speed_input.split(" mph")[0]) * 1.609344
        # If it contains other letters, it can't be parsed as speed.
        if any(c.isalpha() for c in speed_input):
            return 0.0
        # Split on the different separators (";" or "-" or "|")
        # and take the mean of the results (if any).
        return mean(int(x) for x in split(r";|-|\|", speed_input) if x.isnumeric())

    def _get_avg_speed_from_graph(
        self, road_type: list[RoadTypeEnum], avg_speed: AvgSpeed
    ) -> float:
        def get_main_speed(link_type: RoadTypeEnum) -> float:
            _base_rt = RoadTypeEnum.get_enum(str(link_type).split("_link")[0])
            if [_base_rt] in avg_speed.road_types:
                return avg_speed.get_avg_speed([_base_rt])
            return 0.0

        def get_uncombined_speed(combined_types: list[RoadTypeEnum]) -> float:
            for _rt in combined_types:
                if [_rt] in avg_speed.road_types:
                    return avg_speed.get_avg_speed([_rt])
            return 0.0

        # Try to get the speed from the uncombined road types in the list.
        if len(road_type) > 1:
            _uncombined_speed = get_uncombined_speed(road_type)
            if _uncombined_speed:
                return _uncombined_speed

        # For the links take the one of the same type of the main roads.
        if "_link" in str(road_type[0]):
            _main_speed = get_main_speed(road_type[0])
            if _main_speed:
                return _main_speed

        return 0.0

    def _get_avg_speed(
        self,
        road_type_col_name: str,
    ) -> AvgSpeed:
        _avg_speed = AvgSpeed()

        for _rt in set(
            str(edata[road_type_col_name]) for _, _, edata in self.graph.edges.data()
        ):
            edge_data = [
                (self.parse_speed(edata["maxspeed"]), edata["length"])
                for _, _, edata in self.graph.edges.data()
                if (edata[road_type_col_name] == _rt) & ("maxspeed" in edata)
            ]
            if not edge_data:
                _avg_speed.set_avg_speed(AvgSpeed.get_road_type_list(_rt), 0.0)
            else:
                # Calculate weighted average speed
                edge_speed, edge_length = zip(*edge_data)
                _avg_speed.set_avg_speed(
                    AvgSpeed.get_road_type_list(_rt),
                    sum(s * l for s, l in zip(edge_speed, edge_length))
                    / sum(edge_length),
                )

        # For all types without an average speed, take one that is most likely.
        _zero_rts = [
            _rt for _rt in _avg_speed.road_types if _avg_speed.get_avg_speed(_rt) == 0
        ]
        if _zero_rts:
            logging.info(
                "Not all of the edges contain a 'maxspeed' attribute. RA2CE will guess the right average maximum "
                "speed per road type that does not contain a 'maxspeed' attribute. Please check the average speed CSV to ensure correct speeds."
            )
            for _zrt in _zero_rts:
                _speed = self._get_avg_speed_from_graph(_zrt, _avg_speed)
                if _speed:
                    _avg_speed.set_avg_speed(_zrt, _speed)
                    continue

                _avg_speed.set_avg_speed(_zrt, _avg_speed.default_speed)
                logging.warning(
                    "Default speed have been assigned to road type %s. Please check the average speed CSV, "
                    "enter the right average speed for this road type and run RA2CE again.",
                    _zrt,
                )

        return _avg_speed

    def _calculate(self, output_graph_dir: Path | None) -> AvgSpeed:
        """
        Calculates the average speed from OSM roads, per road type.
        If the average speed is already calculated and saved in a CSV, it will be read from there.
        After calculation it will be saved in a CSV file in the output_graph_dir.

        Args:
            output_graph_dir (Path | None): Directory to read/save the average speed CSV file.

        Returns:
            AvgSpeed: Object with the average speeds per road type
        """
        if output_graph_dir:
            _avg_speed_path = output_graph_dir.joinpath("avg_speed.csv")
            if _avg_speed_path.is_file():
                return AvgSpeedReader().read(_avg_speed_path)

            logging.warning(
                "No valid file found with average speeds %s, calculating and saving them instead.",
                _avg_speed_path,
            )

        _length_array, _maxspeed_array = list(
            zip(
                *(
                    ("length" in e, "maxspeed" in e)
                    for _, _, e in self.graph.edges.data()
                )
            )
        )
        if all(_length_array) and any(_maxspeed_array):
            # Add time weighing - Define and assign average speeds; or take the average speed from an existing CSV
            _avg_speed = self._get_avg_speed("highway")
            if output_graph_dir:
                AvgSpeedWriter().export(
                    output_graph_dir.joinpath("avg_speed.csv"), _avg_speed
                )
        else:
            _avg_speed = AvgSpeed()
            logging.info(
                "No attributes found in the graph to estimate average speed per network segment."
            )

        return _avg_speed

    def assign(self) -> nx.Graph:
        """
        Assigns the average speed and time to roads in an existing (OSM) graph.
        If a road does not have a 'maxspeed' attribute, the average speed will be assigned based on the road type.

        Args: None

        Returns:
            graph (NetworkX graph): NetworkX graph with an additional attribute 'avgspeed' and 'time'
        """

        def get_speed(edata: dict) -> float:
            _rt = edata.get("highway", None)
            _speed = self.parse_speed(edata.get("maxspeed", None))
            if not _speed:
                _speed = self.avg_speed.get_avg_speed(AvgSpeed.get_road_type_list(_rt))
            return max(round(_speed, 0), 1)

        for u, v, k, edata in self.graph.edges.data(keys=True):
            _speed = get_speed(edata)  # km/h
            _length = edata.get("length", None)  # m
            if not _length:
                _length = nut.line_length(edata["geometry"], self.graph.graph["crs"])
                self.graph.edges[u, v, k]["length"] = _length
            self.graph.edges[u, v, k]["avgspeed"] = max(
                round(_speed, 0), 1
            )  # km/h (at least 1)
            self.graph.edges[u, v, k]["time"] = round(_length * 1e-3 / _speed, 3)  # h

        return self.graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path

import geopandas as gpd

from ra2ce.network.exporters.network_exporter_base import NetworkExporterBase


class GeoDataFrameNetworkExporter(NetworkExporterBase):
    def export_to_gpkg(self, output_dir: Path, export_data: gpd.GeoDataFrame) -> None:
        _output_shp_path = output_dir / (self._basename + ".gpkg")
        export_data.to_file(
            _output_shp_path, index=False
        )  # , encoding='utf-8' -Removed the encoding type because this causes some shapefiles not to save.
        logging.info(f"Saved {_output_shp_path.stem} in {output_dir}.")

    def export_to_pickle(self, output_dir: Path, export_data: gpd.GeoDataFrame) -> None:
        self.pickle_path = output_dir / (self._basename + ".feather")
        export_data.to_feather(self.pickle_path, index=False)
        logging.info(f"Saved {self.pickle_path.stem} in {output_dir}.")



from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class AnalysisLossesEnum(Ra2ceEnumBase):
    SINGLE_LINK_REDUNDANCY = 1
    MULTI_LINK_REDUNDANCY = 2
    OPTIMAL_ROUTE_ORIGIN_DESTINATION = 3
    MULTI_LINK_ORIGIN_DESTINATION = 4
    OPTIMAL_ROUTE_ORIGIN_CLOSEST_DESTINATION = 5
    MULTI_LINK_ORIGIN_CLOSEST_DESTINATION = 6
    SINGLE_LINK_LOSSES = 7
    MULTI_LINK_LOSSES = 8
    MULTI_LINK_ISOLATED_LOCATIONS = 9
    INVALID = 99



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

import geopandas as gpd
from geopandas import GeoDataFrame
from networkx import MultiGraph

from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_wrappers.network_wrapper_protocol import (
    NetworkWrapperProtocol,
)
from ra2ce.network.networks_utils import graph_from_gdf
from ra2ce.network.segmentation import Segmentation


class TrailsNetworkWrapper(NetworkWrapperProtocol):
    def __init__(self, config_data: NetworkConfigData) -> None:
        logging.info(
            """The original OSM PBF import is no longer supported. 
                Instead, the beta version of package TRAILS is used. 
                First stable release of TRAILS is expected in 2023."""
        )
        self.primary_files = config_data.network.primary_file
        self.segmentation_length = config_data.cleanup.segmentation_length
        self.crs = config_data.crs

    def get_network(self) -> tuple[MultiGraph, GeoDataFrame]:
        """Creates a network which has been prepared in the TRAILS package

        #Todo: we might later simply import the whole trails code as a package, and directly use these functions
        #Todo: because TRAILS is still in beta version we better wait with that untill the first stable version is
        # released

        Returns:
            graph_simple (NetworkX graph): Simplified graph (for use in the losses analyses).
            complex_edges (GeoDataFrame): Complex graph (for use in the damages analyses).
        """
        logging.info(
            "TRAILS importer: Reads the provided primary edge file: {}, assumes there also is a_nodes file".format(
                self.primary_files
            )
        )

        logging.warning(
            "Any coordinate projection information in the feather file will be overwritten (with default WGS84)"
        )
        # Make a pyproj CRS from the EPSG code

        _edge_file = self.primary_files[0]
        edges = gpd.read_feather(_edge_file)
        edges = edges.set_crs(self.crs)

        corresponding_node_file = _edge_file.replace("edges", "nodes")
        if not corresponding_node_file.exists():
            raise FileNotFoundError(
                "The node file could not be found while importing from TRAILS"
            )

        nodes = gpd.read_feather(corresponding_node_file)
        nodes = nodes.set_crs(self.crs)

        logging.info("TRAILS importer: start generating graph")
        # tempfix to rename columns
        edges = edges.rename({"from_id": "node_A", "to_id": "node_B"}, axis="columns")
        node_id = "id"
        graph_simple = graph_from_gdf(edges, nodes, name="network", node_id=node_id)

        logging.info("TRAILS importer: graph generating was succesfull.")
        logging.warning(
            "RA2CE will not clean-up your graph, assuming that it is already done in TRAILS"
        )

        edges_complex = edges
        if self.segmentation_length:
            logging.info("TRAILS importer: start segmentating graph")
            to_segment = Segmentation(edges, self.segmentation_length)
            edges_simple_segmented = to_segment.apply_segmentation()
            if edges_simple_segmented.crs is None:  # The CRS might have disappeared.
                edges_simple_segmented.crs = edges.crs  # set the right CRS
                edges_complex = edges_simple_segmented

        graph_complex = graph_simple  # NOTE THAT DIFFERENCE
        # BETWEEN SIMPLE AND COMPLEX DOES NOT EXIST WHEN IMPORTING WITH TRAILS

        # Todo: better control over metadata in trails
        # Todo: better control over where things are saved in the pipeline
        return graph_complex, edges_complex



from ra2ce.analysis.losses.traffic_analysis.traffic_analysis_base import (
    TrafficAnalysisBase,
)
from ra2ce.analysis.losses.traffic_analysis.traffic_analysis_factory import (
    TrafficAnalysisFactory,
)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Protocol, runtime_checkable

from ra2ce.common.configuration.config_data_protocol import ConfigDataProtocol
from ra2ce.common.io.readers.file_reader_protocol import FileReaderProtocol


@runtime_checkable
class ConfigDataReaderProtocol(FileReaderProtocol, Protocol):  # pragma: no cover
    def read(self, ini_file: Path) -> ConfigDataProtocol:
        """
        Reads the given `*.ini` file and if possible converts it into a `IniConfigDataProtocol` object.

        Args:
            ini_file (Path): Ini file to be mapped into a `IniConfigDataProtocol`.

        Returns:
            ConfigWrapperProtocol: Resulting mapped object from the configuration data in the given file.
        """
        pass



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import logging
from pathlib import Path

import pandas as pd


class MaxDamageByRoadTypeByLane:
    """
    Max damage per RoadType and per Lane

    Attributes:
        self.name (str) : Name of the damage curve
        self.data (pd.DataFrame) : columns contain number of lanes; rows contain the road types

    Optional attributes:
        self.origin_path (Path) : Path to the file from which the function was constructed
        self.raw_data : The raw data read from the input file


    """

    def __init__(self, name=None, damage_unit=None):
        self.name = name
        self.damage_unit = damage_unit

    def from_csv(self, path: Path, sep=",") -> None:
        """Construct object from csv file. Damage curve name is inferred from filename

        The first row describes the lane numbers per column and should have 'Road_type \ lanes' as index/first value.
        Assume road types are in the rows; lane numbers in the columns.
        The second row has the units per column, and should have 'unit' as index/first value
        the rest of the rows contains the different road types as index/first value; and the costs as values

        Arguments:
            *path* (Path) : Path to the csv file
            *sep* (str) : csv seperator
        """
        default_output_unit = "euro/m"
        self.name = path.stem
        self.raw_data = pd.read_csv(path, index_col=r"Road_type \ lanes", sep=sep)
        self.origin_path = path  # to track the original path from which the object was constructed; maybe also date?

        ###Determine units
        units = self.raw_data.loc["unit", :].unique()  # identify the unique units
        assert (
            len(units) == 1
        ), "Columns in the max damage csv seem to have different units, ra2ce cannot handle this"
        # case only one unique unit is identified
        self.damage_unit = units[
            0
        ]  # should have the structure 'x/y' , e.g. euro/m, dollar/yard

        self.data = self.raw_data.drop("unit")
        self.data = self.data.astype("float")

        # assume road types are in the rows; lane numbers in the columns
        self.road_types = list(self.data.index)  # to method
        # assumes that the columns containst the lanes
        self.data.columns = self.data.columns.astype("int")

        if self.damage_unit != default_output_unit:
            self.convert_length_unit()  # convert the unit

    def convert_length_unit(self, desired_unit="euro/m"):
        """Converts max damage values to a different unit
        Arguments:
            self.damage_unit (implicit)
            *desired_unit* (string)

        Effect: converts the values in self.data; and sets the new damage_unit

        """
        if desired_unit == self.damage_unit:
            logging.info("Input damage units are already in the desired format")
            return

        original_length_unit = self.damage_unit.split("/")[1]
        target_length_unit = desired_unit.split("/")[1]

        if original_length_unit != "km" or target_length_unit != "m":
            # We currently only support from 'km' to 'm'
            logging.warning(
                "Damage scaling from {} to {} is not supported".format(
                    self.damage_unit, desired_unit
                )
            )
            return

        scaling_factor = 1 / 1000
        self.data = self.data * scaling_factor
        logging.info(
            "Damage data from {} was scaled by a factor {}, to convert from {} to {}".format(
                self.origin_path, scaling_factor, self.damage_unit, desired_unit
            )
        )
        self.damage_unit = desired_unit



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import shutil
from pathlib import Path
from typing import Optional

from ra2ce.analysis.analysis_config_data.analysis_config_data import AnalysisConfigData
from ra2ce.analysis.analysis_config_data.analysis_config_data_reader import (
    AnalysisConfigDataReader,
)
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.configuration.config_wrapper import ConfigWrapper
from ra2ce.network.network_config_data.network_config_data_reader import (
    NetworkConfigDataReader,
)
from ra2ce.network.network_config_wrapper import NetworkConfigWrapper


class ConfigFactory:
    @staticmethod
    def get_config_wrapper(
        network_ini: Optional[Path], analysis_ini: Optional[Path]
    ) -> ConfigWrapper:
        """
        Generates a `ConfigWrapper` containing the DataObjectModel representations of the given network and analysis ini files.

        Args:
            network_ini (Path): Path to the  `network.ini` file.
            analysis_ini (Path): Path to the `analysis.ini` file.

        Returns:
            ConfigWrapper: Instantiated `ConfigWrapper`.
        """
        _input_config = ConfigWrapper()
        _input_config.network_config = ConfigFactory.get_network_config_data(
            network_ini
        )
        _input_config.analysis_config = ConfigFactory.get_analysis_config_data(
            analysis_ini, _input_config.network_config
        )
        return _input_config

    @staticmethod
    def get_network_config_data(network_ini: Path) -> Optional[NetworkConfigWrapper]:
        if not network_ini:
            return None
        _network_config = NetworkConfigDataReader().read(network_ini)
        # Copy the network ini file to the output directory.
        if not _network_config.output_path.exists():
            _network_config.output_path.mkdir(parents=True)
        _output_ini = _network_config.output_path.joinpath(network_ini.name)
        shutil.copyfile(network_ini, _output_ini)
        return NetworkConfigWrapper.from_data(network_ini, _network_config)

    @staticmethod
    def get_analysis_config_data(
        analysis_ini: Path, network_config: Optional[NetworkConfigWrapper]
    ) -> Optional[AnalysisConfigWrapper]:
        if not analysis_ini:
            return None
        _analysis_config = AnalysisConfigDataReader().read(analysis_ini)
        if isinstance(network_config, NetworkConfigWrapper):
            return ConfigFactory.get_analysis_config_with_network(
                analysis_ini, _analysis_config, network_config
            )
        else:
            return ConfigFactory.get_analysis_config_without_network(
                analysis_ini, _analysis_config
            )

    @staticmethod
    def get_analysis_config_with_network(
        analysis_ini: Path,
        config_data: AnalysisConfigData,
        network_config: NetworkConfigWrapper,
    ):
        return AnalysisConfigWrapper().from_data_with_network(
            analysis_ini, config_data, network_config
        )

    @staticmethod
    def get_analysis_config_without_network(
        analysis_ini: Path,
        config_data: AnalysisConfigData,
    ):
        _network_config = NetworkConfigWrapper()

        # Read existing files and graphs from static folder
        _static_dir = config_data.static_path
        if _static_dir and _static_dir.is_dir():
            _network_config.graph_files = NetworkConfigWrapper.read_graphs_from_config(
                _static_dir.joinpath("output_graph")
            )
        else:
            logging.error(f"Static dir not found. Value provided: {_static_dir}")

        # Read network config file to get network and origin_destination settings
        if config_data.output_path:
            _output_network_ini_file = config_data.output_path.joinpath("network.ini")
        else:
            _output_network_ini_file = Path()
        if _output_network_ini_file.is_file():
            _network_data = NetworkConfigDataReader().read(_output_network_ini_file)
            _network_config.config_data.network = _network_data.network
            _network_config.config_data.origins_destinations = (
                _network_data.origins_destinations
            )
        else:
            logging.error(
                f"Network configuration not found. Value provided: {_output_network_ini_file}"
            )

        return AnalysisConfigWrapper().from_data_with_network(
            analysis_ini, config_data, _network_config
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Protocol, runtime_checkable

from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper
from ra2ce.configuration.config_wrapper import ConfigWrapper


@runtime_checkable
class AnalysisRunner(Protocol):
    @staticmethod
    def can_run(ra2ce_input: ConfigWrapper) -> bool:
        """
        Validates whether the given `ConfigWrapper` is eligibile for this `AnalysisRunner`.

        Args:
            ra2ce_input (ConfigWrapper): Configuration desired to run.

        Returns:
            bool: Whether the `ConfigWrapper` can be run or not.
        """

    def run(
        self, analysis_config: AnalysisConfigWrapper
    ) -> list[AnalysisResultWrapper]:
        """
        Runs this `AnalysisRunner` with the given analysis configuration.

        Args:
            analysis_config (AnalysisConfigWrapper): Analysis configuration representation to be run on this `AnalysisRunner`.

        Returns:
            list[AnalysisResultWrapper]: List of all results for all ran analysis.
        """



from __future__ import annotations

from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class TrafficPeriodEnum(Ra2ceEnumBase):
    NONE = 0
    DAY = 1
    WEEK = 2
    MONTH = 3
    YEAR = 4
    MORNING_PEAK = 5
    EVENING_PEAK = 6
    OFF_PEAK = 7
    NIGHT = 8
    CUSTOM = 9
    INVALID = 99

    @classmethod
    def get_enum(cls, input: str | None) -> TrafficPeriodEnum:
        return super().get_enum(input)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import json
import logging
from pathlib import Path
from typing import Any

from ra2ce.common.io.writers.ra2ce_exporter_protocol import Ra2ceExporterProtocol


class JsonExporter(Ra2ceExporterProtocol):
    def export(self, export_path: Path, export_data: Any) -> None:
        """
        Exports into JSON the given data at the given path. When the parent(s) directory does not exist then it will be created.

        Args:
            export_path (Path): File path where to store the final 'json' file.
            export_data (Any): Data to export.
        """
        _export_dir = export_path.parent
        if not _export_dir.is_dir():
            _export_dir.mkdir(parents=True)

        with open(export_path, "w") as _export_strem:
            json.dump(export_data, _export_strem)
            logging.info(f"Saved (or overwrote) {export_path.name}")



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass, field

from ra2ce.analysis.analysis_config_data.enums.traffic_period_enum import (
    TrafficPeriodEnum,
)
from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum


@dataclass(kw_only=True)
class TrafficIntensities:
    """
    Class to store the traffic intensities per day for different trip types.
    """

    link_id: list[int | tuple[int, int]] = field(default_factory=list)
    intensities: dict[tuple[TrafficPeriodEnum, TripPurposeEnum], list[int]] = field(
        default_factory=dict
    )

    def calculate_intensity(
        self,
        link_id: int | tuple[int, int],
        traffic_period: TrafficPeriodEnum,
        trip_purpose: TripPurposeEnum,
    ) -> int:
        """
        Calculate the traffic intensity per traffic period for a specific link
        for a trip purpose.

        For a simplified graph, the link_id could be a tuple of link_ids.
        In that case the maximum intensity of the links is returned.

        Args:
            link_id (int | tuple[int, int]): The link id(s)
            traffic_period (TrafficPeriodEnum): Part of the day
            trip_purpose (TripPurposeEnum): Trip purpose

        Returns:
            int: The intensity for that (set of) link(s) (vehicles per traffic period)
        """
        if isinstance(link_id, tuple):
            return max(
                self.calculate_intensity(_id, traffic_period, trip_purpose)
                for _id in link_id
            )
        return self.intensities[(traffic_period, trip_purpose)][
            self.link_id.index(link_id)
        ]



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass

import networkx as nx
from snkit.network import Network as SnkitNetwork


@dataclass(kw_only=True)
class SnkitToNxNetworkConverter:
    """
    Class responsible to convert a `snkit.network.Network` into
    a matching `networkx.MultiGraph`.
    """

    snkit_network: SnkitNetwork
    node_id_column_name: str = "id"
    edge_from_id_column_name: str = "from_id"
    edge_to_id_column_name: str = "to_id"

    def convert(self) -> nx.MultiGraph:
        """
        Converts the given `snkit.network.Network` into a matching
        `networkx.MultiGraph`.

        Args:
            snkit_network (SnkitNetwork): The snkit network to convert.

        Returns:
            `networkx.MultiGraph`: The converted graph.
        """
        # Define new graph
        _nx_graph = nx.MultiGraph()
        _crs = self.snkit_network.edges.crs

        # Add nodes to the graph
        for _, row in self.snkit_network.nodes.iterrows():
            node_id = row[self.node_id_column_name]
            attributes = {k: v for k, v in row.items()}
            _nx_graph.add_node(node_id, **attributes)

        # Add edges to the graph
        for _, row in self.snkit_network.edges.iterrows():
            u = row[self.edge_from_id_column_name]
            v = row[self.edge_to_id_column_name]
            attributes = {k: v for k, v in row.items()}
            _nx_graph.add_edge(u, v, **attributes)

        # Add CRS information to the graph
        if "crs" not in _nx_graph.graph:
            _nx_graph.graph["crs"] = _crs

        return _nx_graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Any

from ra2ce.common.validation.ra2ce_validator_protocol import Ra2ceIoValidator
from ra2ce.common.validation.validation_report import ValidationReport
from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase
from ra2ce.network.network_config_data.enums.source_enum import SourceEnum
from ra2ce.network.network_config_data.network_config_data import (
    HazardSection,
    NetworkConfigData,
    NetworkSection,
    ProjectSection,
)

NetworkDictValues: dict[str, list[Any]] = {
    "polygon": ["file", None],
    "directed": [True, False, None],
    "origins": ["file", None],
    "destinations": ["file", None],
    "save_gpkg": [True, False, None],
    "save_csv": [True, False, None],
    "hazard_map": ["file", None],
    "save_traffic": [True, False, None],
    "locations": ["file", None],
}


class NetworkConfigDataValidator(Ra2ceIoValidator):
    def __init__(self, config_data: NetworkConfigData) -> None:
        self._config = config_data

    def _validate_shp_input(self, network_config: NetworkSection) -> ValidationReport:
        """Checks if a file id is configured when using the option to create network from shapefile"""
        _shp_report = ValidationReport()
        if network_config.source == SourceEnum.SHAPEFILE and not network_config.file_id:
            _shp_report.error(
                "Not possible to create network - Shapefile used as source, but no file_id configured in the network.ini file"
            )
        return _shp_report

    def validate(self) -> ValidationReport:
        """Check if input properties are correct and exist."""
        _report = ValidationReport()
        if not self._config.network:
            _report.error("Network properties not present in Network ini file.")
            return _report

        # check if properties exist in settings.ini file
        _report.merge(self._validate_shp_input(self._config.network))
        _report.merge(self._validate_sections())
        return _report

    def _wrong_value(self, key: str) -> str:
        _accepted_values = ", ".join(NetworkDictValues[key])
        return (
            f"Wrong input to property [ {key} ]; has to be one of: {_accepted_values}."
        )

    def _wrong_enum(self, key: str, enum: Ra2ceEnumBase) -> str:
        # Remove last value INVALID
        _accepted_values = enum.list_valid_options()
        return (
            f"Wrong input to property [ {key} ]; has to be one of: {_accepted_values}."
        )

    def _validate_enum(self, enum: Ra2ceEnumBase, key: str) -> ValidationReport:
        _report = ValidationReport()
        if not enum.is_valid():
            _report.error(self._wrong_enum(key, enum))
        return _report

    def _validate_project_section(
        self, project_section: ProjectSection
    ) -> ValidationReport:
        _report = ValidationReport()
        if not project_section:
            _report.error("Project section not found.")
        return _report

    def _validate_network_section(
        self, network_section: NetworkSection
    ) -> ValidationReport:
        _network_report = ValidationReport()

        # Validate source
        _network_report.merge(self._validate_enum(network_section.source, "source"))

        # Validate network_type
        _network_report.merge(
            self._validate_enum(network_section.network_type, "network_type")
        )

        # Validate road types.
        if network_section.road_types and any(
            not _type.is_valid() for _type in network_section.road_types
        ):
            _network_report.error(
                f"Wrong road type(s) configured; has to be one or multiple of: {[_type.config_value for _type in network_section.road_types[0].list_valid_options()]}"
            )
        return _network_report

    def _validate_hazard_section(
        self, hazard_section: HazardSection
    ) -> ValidationReport:
        _hazard_report = ValidationReport()

        _hazard_report.merge(
            self._validate_enum(hazard_section.aggregate_wl, "aggregate_wl")
        )

        return _hazard_report

    def _validate_sections(self) -> ValidationReport:
        _report = ValidationReport()
        _available_keys = self._config.__dict__.keys()
        _required_sections = [
            "project",
            "network",
            "origins_destinations",
            "hazard",
            "cleanup",
        ]
        for _required_section in _required_sections:
            if _required_section not in _available_keys:
                _report.error(
                    f"Section [ {_required_section} ] is not configured. Add section [ {_required_section} ] to the *.ini file. "
                )

        if not _report.is_valid():
            return _report

        _report.merge(self._validate_project_section(self._config.project))
        _report.merge(self._validate_network_section(self._config.network))
        _report.merge(self._validate_hazard_section(self._config.hazard))

        return _report



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Optional

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_protocol import AnalysisProtocol
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames


class AnalysisLossesProtocol(AnalysisProtocol):
    graph_file: Optional[GraphFile]
    graph_file_hazard: Optional[GraphFile]
    analysis: AnalysisSectionLosses
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path

from networkx import Graph
from osgeo import gdal

from ra2ce.network.networks_utils import bounds_intersect_2d, get_extent


def validate_extent_graph(extent_graph: list[float], tif_hazard_file: Path) -> None:
    """
    Validates the given extent graph to a hazard file (*.tif)

    Args:
        extent_graph (list[float]): List of boundary points determening the extent of a graph.
        tif_hazard_file (Path): Hazard (*.tif) file.

    Raises:
        ValueError: When the hazard raster and the graph geometries do not overlap.
    """
    # Check if the hazard and graph extents overlap
    extent = get_extent(gdal.Open(str(tif_hazard_file)))
    extent_hazard = (
        extent["minX"],
        extent["maxX"],
        extent["minY"],
        extent["maxY"],
    )

    if not bounds_intersect_2d(extent_graph, extent_hazard):
        logging.info("Raster extent: {}, Graph extent: {}".format(extent, extent_graph))
        raise ValueError(
            "The hazard raster and the graph geometries do not overlap, check projection"
        )


def get_edges_geoms(graph: Graph) -> list:
    """
    Gets all edges geometry from a provided graph.
    """
    return [
        (u, v, k, edata)
        for u, v, k, edata in graph.edges.data(keys=True)
        if "geometry" in edata
    ]



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from ast import literal_eval
from collections import defaultdict
from dataclasses import dataclass, field

from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


@dataclass
class RoadTypeEntry:
    road_type: list[RoadTypeEnum]

    def __str__(self) -> str:
        """
        Override the str function to make it writeable as key in the CSV.
        """
        if len(self.road_type) == 1:
            return self.road_type[0].config_value
        return str([x.config_value for x in self.road_type])

    def __hash__(self) -> int:
        """
        Override the hash function to make the RoadTypeEntry hashable.
        """
        return hash(str(self.road_type))


@dataclass
class AvgSpeed:
    default_speed: float = 50.0
    speed_per_road_type: defaultdict[RoadTypeEntry, float] = field(
        default_factory=lambda: defaultdict(lambda: AvgSpeed.default_speed)
    )

    @property
    def road_types(self) -> list[list[RoadTypeEnum]]:
        return [_rte.road_type for _rte in self.speed_per_road_type.keys()]

    @staticmethod
    def get_road_type_list(
        road_type: str | list[str] | None,
    ) -> list[RoadTypeEnum] | None:
        if not road_type:
            return [RoadTypeEnum.INVALID]
        if isinstance(road_type, str):
            if road_type.startswith("["):
                # If the roadtype is a str(list), convert it back to a list
                return list(map(RoadTypeEnum.get_enum, literal_eval(road_type)))
            return [RoadTypeEnum.get_enum(road_type)]
        if isinstance(road_type, list):
            return list(map(RoadTypeEnum.get_enum, road_type))
        else:
            return [RoadTypeEnum.INVALID]

    def get_avg_speed(self, road_type: list[RoadTypeEnum]) -> float:
        return self.speed_per_road_type[RoadTypeEntry(road_type)]

    def set_avg_speed(self, road_type: list[RoadTypeEnum], avg_speed: float) -> None:
        self.speed_per_road_type[RoadTypeEntry(road_type)] = round(avg_speed, 1)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass, field

from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_factory import AnalysisFactory
from ra2ce.analysis.damages.analysis_damages_protocol import AnalysisDamagesProtocol
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol


@dataclass
class AnalysisCollection:
    damages_analyses: list[AnalysisDamagesProtocol] = field(default_factory=list)
    losses_analyses: list[AnalysisLossesProtocol] = field(default_factory=list)

    @classmethod
    def from_config(cls, analysis_config: AnalysisConfigWrapper) -> AnalysisCollection:
        """
        Create an AnalysisCollection from an AnalysisConfigWrapper.

        Args:
            analysis_config (AnalysisConfigWrapper): The analysis configuration.

        Returns:
            AnalysisCollection: Collection of analyses to be executed.
        """
        return cls(
            damages_analyses=[
                AnalysisFactory.get_damages_analysis(analysis, analysis_config)
                for analysis in analysis_config.config_data.damages_list
            ],
            losses_analyses=[
                AnalysisFactory.get_losses_analysis(analysis, analysis_config)
                for analysis in analysis_config.config_data.losses_list
            ],
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Union


@dataclass
class AccumulatedTraffic:
    utilitarian: float = 0.0
    egalitarian: float = 0.0
    prioritarian: float = 0.0

    def __add__(
        self, other: Union[AccumulatedTraffic, float, int]
    ) -> AccumulatedTraffic:
        """
        Overloading of the add (`operator.add` or simply `+`).
        This overload ensures we correctly add the `AccumulatedTraffic` attributes (`utilitarian`, `egalitarian`, `prioritarian`).
        """
        if isinstance(other, AccumulatedTraffic):
            return AccumulatedTraffic(
                utilitarian=self.utilitarian + other.utilitarian,
                prioritarian=self.prioritarian + other.prioritarian,
                egalitarian=self.egalitarian + other.egalitarian,
            )
        elif isinstance(other, float) or isinstance(other, int):
            return AccumulatedTraffic(
                utilitarian=self.utilitarian + other,
                prioritarian=self.prioritarian + other,
                egalitarian=self.egalitarian + other,
            )
        raise TypeError(
            "It is not possible to sum {} with a value of type {}.".format(
                AccumulatedTraffic.__name__, type(other).__name__
            )
        )

    def __mul__(
        self, other: Union[AccumulatedTraffic, float, int]
    ) -> AccumulatedTraffic:
        """
        Overloading of the multiply (`operator.mul` or simply `*`).
        This overload ensures we correctly multiply the `AccumulatedTraffic` attributes (`utilitarian`, `egalitarian`, `prioritarian`).
        """
        if isinstance(other, AccumulatedTraffic):
            return AccumulatedTraffic(
                utilitarian=self.utilitarian * other.utilitarian,
                prioritarian=self.prioritarian * other.prioritarian,
                egalitarian=self.egalitarian * other.egalitarian,
            )
        elif isinstance(other, float) or isinstance(other, int):
            return AccumulatedTraffic(
                utilitarian=self.utilitarian * other,
                prioritarian=self.prioritarian * other,
                egalitarian=self.egalitarian * other,
            )
        raise TypeError(
            "It is not possible to multiply {} with a value of type {}.".format(
                AccumulatedTraffic.__name__, type(other).__name__
            )
        )



from pathlib import Path
from typing import Protocol, runtime_checkable

from geopandas import GeoDataFrame
from networkx import MultiGraph


@runtime_checkable
class GraphFileProtocol(Protocol):
    name: str
    folder: Path
    graph: MultiGraph | GeoDataFrame

    @property
    def file(self) -> Path | None:
        """
        Return the path to the graph

        Returns:
            Path | None: _description_
        """

    def read_graph(self, folder: Path) -> None:
        """
        Read a graph file

        Args:
            folder (Path): Folder of the graph

        Returns: None
        """
        pass

    def get_graph(self) -> MultiGraph | GeoDataFrame:
        """
        Gets a graph file that is already known in the object.
        It is read if not done yet.

        Args: None

        Returns:
            MultiGraph | GeoDataFrame: the graph
        """
        pass



from typing import Any

from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.analysis.losses.weighing_analysis.weighing_analysis_protocol import (
    WeighingAnalysisProtocol,
)


class TimeWeighingAnalysis(WeighingAnalysisProtocol):
    time_list: list
    edge_data: dict[str, Any]

    def __init__(self) -> None:
        self.time_list = []
        self.edge_data = {}

    def _calculate_time(self, dist: float) -> float:
        _avgspeed = self.edge_data.get("avgspeed", None)  # km/h
        return round(dist * 1e-3 / _avgspeed, 3)  # h

    def get_current_value(self) -> float:
        _time = self.edge_data.get(WeighingEnum.TIME.config_value, None)  # h
        if not _time:
            _dist = self.edge_data.get(WeighingEnum.LENGTH.config_value, 0)  # m
            _time = self._calculate_time(_dist)
            self.edge_data[WeighingEnum.TIME.config_value] = _time
        self.time_list.append(_time)
        return _time



import copy
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, overlay, read_feather
from pyproj import CRS

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.networks_utils import buffer_geometry, graph_to_gdf


class MultiLinkIsolatedLocations(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file_hazard: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names

    def utm_crs(self, bbox: list[float]) -> CRS:
        """Returns wkt string of nearest UTM projects
        Parameters
        ----------
        bbox : array-like of floats
            (xmin, ymin, xmax, ymax) bounding box in latlon WGS84 (EPSG:4326) coordinates
        Returns
        -------
        crs: pyproj.CRS
            CRS of UTM projection

        FROM HYDROMT: https://github.com/Deltares/hydromt - 10.5281/zenodo.6107669
        """
        left, bottom, right, top = bbox
        x = (left + right) / 2
        y = (top + bottom) / 2
        kwargs = dict(zone=int(np.ceil((x + 180) / 6)))
        # BUGFIX hydroMT v0.3.5: south=False doesn't work only add south=True if y<0
        if y < 0:
            kwargs.update(south=True)
        # BUGFIX hydroMT v0.4.6: add datum
        epsg = CRS(proj="utm", datum="WGS84", ellps="WGS84", **kwargs).to_epsg()
        return CRS.from_epsg(epsg)

    def remove_edges_from_largest_component(self, disconnected_graph: nx.Graph) -> None:
        """
        This function removes all edges from the largest connected component of a graph.

        Args:
            disconnected_graph (nx.Graph): The graph from which to remove the edges.
        """
        connected_components = list(
            c for c in nx.connected_components(disconnected_graph)
        )
        connected_components_size = list(
            len(c) for c in nx.connected_components(disconnected_graph)
        )

        largest_comp_index = connected_components_size.index(
            max(connected_components_size)
        )
        edges_from_lagest_component = list(
            disconnected_graph.subgraph(
                connected_components[largest_comp_index]
            ).edges()
        )
        disconnected_graph.remove_edges_from(edges_from_lagest_component)

    def get_network_with_edge_fid(self, graph: nx.Graph) -> GeoDataFrame:
        """
        This function converts a NetworkX graph into a GeoDataFrame representing the network.
        It also constructs an 'edge_fid' column based on the 'node_A' and 'node_B' columns, following a specific convention.

        Args:
            G (nx.Graph): The NetworkX graph to be converted.

        Returns:
            gpd.GeoDataFrame: The resulting GeoDataFrame, with an added 'edge_fid' column.
        """
        network = graph_to_gdf(graph)[0]
        # TODO: add making "edges_fid" (internal convention) to graph_to_gdf
        if all(c_idx in network.index.names for c_idx in ["u", "v"]):
            network["edge_fid"] = [f"{na}_{nb}" for (na, nb, _) in network.index]
        elif all(c_idx in network.columns for c_idx in ["node_A", "node_B"]):
            # shapefiles
            network["edge_fid"] = [
                f"{na}_{nb}" for na, nb in network["node_A", "node_B"].values
            ]
        return network[["edge_fid", "geometry"]]

    def _summarize_locations(
        self, locations: GeoDataFrame, cat_col: str, hazard_id: str
    ) -> pd.DataFrame:
        """
        This function summarizes the hazard impacts on different categories of locations.
        It adds a counter for each hazard type and aggregates the data per category.

        Args:
            locations (gpd.GeoDataFrame): A GeoDataFrame containing the locations and their hazard impacts.
            cat_col (str): The column name in the locations GeoDataFrame that represents the location categories.
            hazard_id (str): The identifier for the hazard being considered.

        Returns:
            pd.DataFrame: A DataFrame summarizing the number of isolated locations per category for the given hazard.
        """
        # add counter
        locations[f"i_{hazard_id}"] = 1  # add counter
        # make an overview of the locations, aggregated per category
        df_aggregation = (
            locations.groupby(by=[cat_col, f"i_type_{hazard_id}"])[f"i_{hazard_id}"]
            .sum()
            .reset_index()
        )
        df_aggregation["hazard"] = hazard_id
        df_aggregation.rename(columns={f"i_{hazard_id}": "nr_isolated"}, inplace=True)
        return df_aggregation

    def multi_link_isolated_locations(
        self, graph: nx.Graph, analysis: AnalysisSectionLosses, crs=4326
    ) -> tuple[GeoDataFrame, pd.DataFrame]:
        """
        This function identifies locations that are flooded or isolated due to the disruption of the network caused by a hazard.
        It iterates over multiple hazard scenarios, modifies the graph to represent direct and indirect impacts, and then
        spatially joins the impacted network with the location data to find out which locations are affected.

        Args:
            graph (nx.Graph): The original graph representing the network, with additional hazard information.
            analysis (AnalysisSectionLosses): The configuration of the analysis, which contains the threshold for considering a hazard impact significant.
            crs (int, optional): The coordinate reference system used for geographical data. Defaults to 4326 (WGS84).

        Returns:
            tuple (gpd.GeoDataFrame, pd.DataFrame): A tuple containing the location GeoDataFrame updated with hazard impacts,
                and a DataFrame summarizing the impacts per location category.
        """

        def _is_not_none(value):
            return (
                value is not None
                and value is not pd.NA
                and not pd.isna(value)
                and not np.isnan(value)
            )

        # Load the point shapefile with the locations of which the isolated locations should be identified.
        locations = read_feather(
            self.static_path.joinpath("output_graph", "locations_hazard.feather")
        )
        # TODO PUT CRS IN DOCUMENTATION OR MAKE CHANGABLE
        # reproject the datasets to be able to make a buffer in meters
        nearest_utm = self.utm_crs(locations.total_bounds)

        # create an empty list to append the df_aggregation to
        aggregation = pd.DataFrame()
        for hazard in self.hazard_names.names:
            # for each hazard event
            hazard_name = self.hazard_names.get_name(hazard)

            graph_hz_direct = copy.deepcopy(graph)
            graph_hz_indirect = copy.deepcopy(graph)

            # filter graph edges that are directly disrupted by the hazard(s), i.e. flooded
            edges = [e for e in graph.edges.data(keys=True) if hazard_name in e[-1]]
            edges_hz_direct = [
                e
                for e in edges
                if (hazard_name in e[-1])
                and (
                    _is_not_none(e[-1][hazard_name])
                    and (e[-1][hazard_name] > float(analysis.threshold))
                    & (
                        ("bridge" not in e[-1])
                        or ("bridge" in e[-1] and e[-1]["bridge"] != "yes")
                    )
                )
            ]
            edges_hz_indirect = [e for e in edges if e not in edges_hz_direct]

            # get indirect graph - remove the edges that are impacted by hazard directly
            graph_hz_indirect.remove_edges_from(edges_hz_direct)
            # get indirect graph without the largest component, i.e. isolated graph
            self.remove_edges_from_largest_component(graph_hz_indirect)

            # get direct graph - romove the edges that are impacted by hazard indirectly
            graph_hz_direct.remove_edges_from(edges_hz_indirect)

            # get isolated network
            network_hz_indirect = GeoDataFrame()
            if len(graph_hz_indirect.edges) > 0:
                network_hz_indirect = self.get_network_with_edge_fid(graph_hz_indirect)
                network_hz_indirect[f"i_type_{hazard_name[:-3]}"] = "isolated"
                # reproject the datasets to be able to make a buffer in meters
                network_hz_indirect = network_hz_indirect.set_crs(crs=crs)
                network_hz_indirect.to_crs(crs=nearest_utm, inplace=True)

            # get flooded network
            network_hz_direct = GeoDataFrame()
            if len(graph_hz_direct.edges) > 0:
                network_hz_direct = self.get_network_with_edge_fid(graph_hz_direct)
                network_hz_direct[f"i_type_{hazard_name[:-3]}"] = "flooded"
                # reproject the datasets to be able to make a buffer in meters
                network_hz_direct = network_hz_direct.set_crs(crs=crs)
                network_hz_direct.to_crs(crs=nearest_utm, inplace=True)

            # get hazard roads
            # merge buffer and set original crs
            results_hz_roads = GeoDataFrame(
                pd.concat([network_hz_direct, network_hz_indirect])
            )
            results_hz_roads = buffer_geometry(
                results_hz_roads, analysis.buffer_meters
            ).to_crs(crs=crs)
            # Save the output
            results_hz_roads.to_file(
                self.output_path.joinpath(
                    analysis.analysis.config_value,
                    f"flooded_and_isolated_roads_{hazard_name}.gpkg",
                )
            )

            # relate the locations to network disruption due to hazard by spatial overlay
            results_hz_roads.reset_index(inplace=True)
            locations_hz = overlay(
                locations, results_hz_roads, how="intersection", keep_geom_type=True
            )

            # Replace nan with 0 for the water depth columns
            # TODO: this should always be done in hazard class
            locations_hz[hazard_name] = locations_hz[hazard_name].fillna(0)

            # TODO: Put in analyses.ini file a variable to set the threshold for locations that are not isolated when they are flooded.
            # Extract the flood depth of the locations
            # intersect = intersect.loc[intersect[hazard_name] > analysis.threshold_locations]

            # get location stats
            df_aggregation = self._summarize_locations(
                locations_hz,
                cat_col=analysis.category_field_name,
                hazard_id=hazard_name[:-3],
            )

            # add to exisiting results
            aggregation = pd.concat([aggregation, df_aggregation], axis=0)

        # Set the locations_hz geopandas dataframe back to the original crs
        locations_hz.to_crs(crs=crs, inplace=True)

        return locations_hz, aggregation

    def execute(self) -> GeoDataFrame:
        _output_path = self.output_path.joinpath(self.analysis.analysis.config_value)

        (gdf, df) = self.multi_link_isolated_locations(
            self.graph_file_hazard.get_graph(), self.analysis
        )

        df_path = _output_path.joinpath(
            self.analysis.name.replace(" ", "_") + "_results.csv"
        )
        df.to_csv(df_path, index=False)

        return gdf



from dataclasses import dataclass
from pathlib import Path

from networkx import MultiGraph

from ra2ce.common.io.readers.graph_pickle_reader import GraphPickleReader
from ra2ce.network.graph_files.graph_files_protocol import GraphFileProtocol


@dataclass
class GraphFile(GraphFileProtocol):
    """
    Note this class resembles NetworkFile to a large extent
    """

    name: str = ""
    folder: Path = None
    graph: MultiGraph = None

    @property
    def file(self) -> Path | None:
        if not self.folder:
            return None
        return self.folder.joinpath(self.name)

    def read_graph(self, folder: Path) -> None:
        if not folder:
            return
        _file = folder.joinpath(self.name)
        if _file and _file.is_file():
            self.folder = folder
            _pickle_reader = GraphPickleReader()
            self.graph = _pickle_reader.read(self.file)

    def get_graph(self) -> MultiGraph:
        if self.graph is None:
            self.read_graph(self.folder)
        return self.graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from decimal import Decimal

import geopandas as gpd
from shapely.geometry import LineString, MultiLineString, Point

from ra2ce.network.networks_utils import cut as network_cut


class Segmentation:  # Todo: more naturally, this would be METHOD of the network class.
    """cut the edges in the complex geodataframe to segments of equal lengths or smaller
    # output will be the cut edges_complex

    Variables:
        *self.edges_input* (Geopandas DataFrame) : the edges that are to be segmented
        *self.segmentation_length* (float) : segmentation lenght in degrees #Todo also in meters?
        *self.save_files* (Boolean) : save segmented graph?

    Result:
        *self.edges_segmented* (Geopandas DataFrame) : the segmented edges dataframe


    """

    def __init__(self, edges_input, segmentation_length, save_files: bool = False):
        # General
        self.edges_input = edges_input  # Edges GeoDataFrame
        self.edges_segmented = (
            None  # This is where the result will be saved Edges GeoDataframe
        )
        self.segmentation_length = segmentation_length
        self.save_files = save_files  # Todo not implemented yet

    def apply_segmentation(self):
        self.cut_gdf()
        logging.info(
            "Finished segmenting the geodataframe with split length: {} degree".format(
                self.segmentation_length
            )
        )
        return self.edges_segmented

    def cut(self, line: LineString, distance: float) -> list:
        """Cuts a line in two at a distance from its starting point

        Args:
            line (LineString): Single linestring.
            distance (float): Distance from starting point of linestring

        Returns:
            (list): A list containing two shapely linestring objects.
        """

        if distance <= 0.0 or distance >= line.length:
            return [LineString(line)]
        coords = list(line.coords)
        for i, p in enumerate(coords):
            pd = line.project(Point(p))
            if pd == distance:
                return [LineString(coords[: i + 1]), LineString(coords[i:])]
            if pd > distance:
                cp = line.interpolate(distance)
                return [
                    LineString(coords[:i] + [(cp.x, cp.y)]),
                    LineString([(cp.x, cp.y)] + coords[i:]),
                ]

    def check_divisibility(self, dividend, divisor):
        """Checks if the dividend is a multiple of the divisor and outputs a boolean value

        Args:
            dividend (float): The number which is divided
            divisor (float): The number which divides

        Returns:
            is_multiple (bool): True if the dividend is a multiple of the divisor, False if not
        """

        dividend = Decimal(str(dividend))
        divisor = Decimal(str(divisor))
        remainder = dividend % divisor

        return remainder == Decimal("0")

    def number_of_segments(self, linestring: LineString, split_length: float) -> int:
        """Returns the integer number of segments which will result from chopping up a linestring with split_length

        Args:
            linestring (LineString): Single linestring.
            split_length (float): The length by which to divide the linestring object.

        Returns:
            n (int): Integer number of segments which will result from splitting linestring with split_length.
        """

        divisible = self.check_divisibility(linestring.length, split_length)
        if divisible:
            return int(linestring.length / split_length)
        return int(linestring.length / split_length) + 1

    def split_linestring(self, linestring: LineString, split_length: float):
        """Cuts a linestring in equivalent segments of length split_length

        Args:
            linestring (LineString): Single linestring.
            split_length (float): Length by which to split the linestring into equal segments.

        Returns:
            result_list (list): List of LineString objects that all have the same length split_lenght.
        """

        n_segments = self.number_of_segments(linestring, split_length)
        if n_segments != 1:
            result_list = [None] * n_segments
            current_right_linestring = linestring

            for i in range(0, n_segments - 1):
                r = network_cut(current_right_linestring, split_length)
                # Can accidently return Nonetypes
                if r is not None:
                    current_left_linestring = r[0]
                    current_right_linestring = r[1]
                    result_list[i] = current_left_linestring
                    result_list[i + 1] = current_right_linestring
                # Sometimes the remainder is so small that it is only one point, which cannot be cut, in that case
                # just pass #Todo: maybe we can do something here to avoid this error

        else:
            result_list = [linestring]

        # Make sure this  function does not return any None objects, because these will cause problems later
        result_list = [
            x
            for x in result_list
            if (type(x) == LineString or type(x) == MultiLineString)
        ]

        return result_list

    def cut_gdf(self):
        """
        Cuts every linestring or multilinestring feature in a gdf to equal length segments. Assumes only linestrings for now.

            *gdf* (GeoDataFrame) : GeoDataFrame to split
            *length* (units of the projection) : Typically in degrees, 0.001 degrees ~ 111 m in Europe
        """
        gdf = self.edges_input.copy()
        columns = gdf.columns

        data = {"splt_id": []}

        for column in columns:
            data[column] = []

        count = 0
        for _, row in gdf.iterrows():
            geom = row["geometry"]
            assert type(geom) == LineString or type(geom) == MultiLineString
            linestrings = self.split_linestring(geom, self.segmentation_length)

            for _, linestring in enumerate(linestrings):
                for key, value in row.items():
                    if key == "geometry":
                        data[key].append(linestring)
                    else:
                        data[key].append(value)
                data["splt_id"].append(count)
                count += 1
        self.edges_segmented = gpd.GeoDataFrame(data)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import logging
from typing import Any, List

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame


def clean_lane_data(lane_col: pd.Series) -> pd.Series:
    """
    Function that cleans up the lane data, to be used in case this contains unexpected values

    @author: Kees van Ginkel, Deltares

    Arguments:
        *lane_col* (Pandas Series) : The Panda series containing all lane information
                                    each value is still a string
    Returns:
        *new_lane_col* (Panda Series) : idem, but cleaned
                                    each value is already a float
    """

    # Todo: drawback of this approach is that you cannot easily see which lanes have the erratic lanedata,
    # The upside is that it is probably faster...
    # for index, cell in lane_col.iteritems():
    #    print(cell, type(cell))
    new_lane_col = lane_col.apply(lambda x: lane_cleaner(x))

    return new_lane_col


def lane_cleaner(cell: Any) -> float:
    """
    Helper function to clean an object with lane data and return it as a float

    @author: Kees van Ginkel, Deltares

    Arguments:
        *cell* (unknown format) : The cell with lane information to be cleaned

    Returns:
        *new* (float) : number of lanes or np.nan
    """
    if cell is None:
        return np.nan
    if isinstance(cell, int):
        return float(cell)
    if isinstance(cell, float):
        return cell
    if isinstance(cell, str):  # try to unpack the cell
        try:
            return float(cell)
        except Exception:
            logging.warning(
                "Lanedata {} could not be converted to float, if it is a list we will try to unpack".format(
                    cell
                )
            )

            def _get_max(list_values: List[str]) -> float:
                try:
                    _max_value = max(map(float, list_values))
                    logging.warning(
                        "Our best guess of the lane number is: {}".format(_max_value)
                    )
                except Exception:
                    logging.warning(
                        "Unexpected datatype, lane data removed {} {}".format(
                            cell, type(cell)
                        )
                    )
                    return np.nan
                return _max_value

            if ";" in cell:  # it looks some sort of a list
                return _get_max(cell.split(";"))
            elif "," in cell:  # it looks some sort of a list
                return _get_max(cell.split(","))
            return np.nan

    logging.warning(
        "Unexpected datatype, lane data removed {} {}".format(cell, type(cell))
    )
    return np.nan


def create_summary_statistics(gdf: GeoDataFrame) -> dict:
    """
    Return the mode (most frequent) #lanes of the available road types in the data

    @author: Kees van Ginkel, Deltares

    Arguments:
        *gdf* (GeoDataFrame) : needs the columns 'road_type' and 'lanes'

    Returns:
        *dictionary* (Dict) : keys = road types; values = lanes

    """
    # Todo: in the future we can make it more generic, so that we can easily get the mode/mean/whatever

    _grouped_lanes = gdf.groupby("road_type")["lanes"]
    _road_types, _lanes = list(zip(*((x[0], x[1].mode()) for x in _grouped_lanes)))
    _lanes_dict = {
        _road_type: _lanes[0] if not _lanes.empty else np.nan
        for _road_type, _lanes in zip(_road_types, _lanes)
    }

    # get a default value if any key of the dictionary is nan
    # (because the mode operation on the 'lanes' column for a road type results in an empty array)
    default_value = np.mean(
        list(_val for _val in _lanes_dict.values() if not np.isnan(_val))
    )

    # Replace nan with the calculated average
    return {
        _road_type: _lanes if not np.isnan(_lanes) else default_value
        for _road_type, _lanes in _lanes_dict.items()
    }


def scale_damage_using_lanes(lane_scale_factors, df, cols_to_scale) -> pd.DataFrame:
    """
    Scale (max) damage or construction cost data using the lane data

    Arguments:
        *lane_scale_factors* (dict)  : output of damages_lookup.get_max_damages_OSD()
        *df* (pd.DataFrame)          : the data to scale, should have a col road_types, and a col lanes
        *cols_to_scale* (list)       : names of the columns to scale

    Returns:
        *df* (pd.Dataframe) : the scaled data
    """
    assert "road_type" in df.columns, "Road type data is missing"
    assert "lanes" in df.columns, "Lane number data is missing"
    df["road_type_scale_factors"] = df.apply(
        lambda x: lane_scale_factors[x.road_type][x.lanes], axis=1
    )

    for col in cols_to_scale:
        df[col] = df[col] * df["road_type_scale_factors"]

    df = df.drop(columns=["road_type_scale_factors"])

    return df



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from pathlib import Path
from typing import Optional, Protocol, runtime_checkable

from ra2ce.common.configuration.config_data_protocol import ConfigDataProtocol
from ra2ce.network.graph_files.graph_files_collection import GraphFilesCollection


@runtime_checkable
class ConfigWrapperProtocol(Protocol):  # pragma: no cover
    ini_file: Path
    config_data: Optional[ConfigDataProtocol] = None
    graph_files: Optional[GraphFilesCollection] = None

    @classmethod
    def from_data(
        cls, ini_file: Path, config_data: ConfigDataProtocol
    ) -> ConfigWrapperProtocol:
        """
        Initializes a `ConfigWrapperProtocol` with the given parameters.

        Args:
            ini_file (Path): Path to the ini file containing the analysis data.
            config_data (IniConfigDataProtocol): Ini data representation.

        Returns:
            ConfigWrapperProtocol: Initialized instance.
        """
        raise NotImplementedError()

    def configure(self) -> None:
        raise NotImplementedError()

    def is_valid(self) -> bool:
        raise NotImplementedError()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import logging
import pickle
from pathlib import Path
from typing import Any

from ra2ce.common.io.readers.file_reader_protocol import FileReaderProtocol


class GraphPickleReader(FileReaderProtocol):
    def read(self, pickle_path: Path) -> Any:
        _read_graph = None
        if not pickle_path:
            raise ValueError("No pickle path was provided.")

        if not pickle_path.is_file():
            _error_mssg = f"No pickle found at path {pickle_path}"
            logging.error(_error_mssg)
            raise ValueError(_error_mssg)

        with open(pickle_path, "rb") as f:
            _read_graph = pickle.load(f)

        return _read_graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass
from pathlib import Path

from pandas import DataFrame

from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum
from ra2ce.analysis.losses.losses_input_data_reader_base import (
    LossesInputDataReaderBase,
)
from ra2ce.analysis.losses.time_values.time_values import TimeValues


@dataclass
class TimeValuesReader(LossesInputDataReaderBase):
    """
    Class to read the time values from a csv file.
    """

    csv_columns = ["trip_types", "value_of_time", "occupants"]
    data_type = TimeValues

    def _parse_df(self, df: DataFrame) -> TimeValues:
        _time_values = {
            TripPurposeEnum.get_enum(_trip_type): (_value_of_time, _occupants)
            for (_trip_type, _value_of_time, _occupants) in zip(
                df["trip_types"], df["value_of_time"], df["occupants"]
            )
        }

        return TimeValues(time_values=_time_values)

    def read(self, file_path: Path | None) -> TimeValues:
        return super().read(file_path)



from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class RoadTypeEnum(Ra2ceEnumBase):
    NONE = 0
    MOTORWAY = 1
    MOTORWAY_LINK = 2
    TRUNK = 3
    TRUNK_LINK = 4
    PRIMARY = 5
    PRIMARY_LINK = 6
    SECONDARY = 7
    SECONDARY_LINK = 8
    TERTIARY = 9
    TERTIARY_LINK = 10
    RESIDENTIAL = 11
    ROAD = 12
    TUNNEL = 13
    BRIDGE = 14
    CULVERT = 15
    UNCLASSIFIED = 98
    INVALID = 99

    @classmethod
    def get_enum(cls, input: str | None) -> Ra2ceEnumBase:
        return super().get_enum(input)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from configparser import ConfigParser
from pathlib import Path
from shutil import copyfile

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisConfigData,
    AnalysisSectionBase,
    AnalysisSectionDamages,
    AnalysisSectionLosses,
    DamagesAnalysisNameList,
    LossesAnalysisNameList,
    ProjectSection,
)
from ra2ce.analysis.analysis_config_data.enums.analysis_damages_enum import (
    AnalysisDamagesEnum,
)
from ra2ce.analysis.analysis_config_data.enums.analysis_losses_enum import (
    AnalysisLossesEnum,
)
from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.analysis_config_data.enums.event_type_enum import EventTypeEnum
from ra2ce.analysis.analysis_config_data.enums.loss_type_enum import LossTypeEnum
from ra2ce.analysis.analysis_config_data.enums.risk_calculation_mode_enum import (
    RiskCalculationModeEnum,
)
from ra2ce.analysis.analysis_config_data.enums.traffic_period_enum import (
    TrafficPeriodEnum,
)
from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum
from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.common.configuration.ini_configuration_reader_protocol import (
    ConfigDataReaderProtocol,
)


class AnalysisConfigDataReader(ConfigDataReaderProtocol):
    """
    Reads the analysis configuration file analyses.ini.
    Takes network_data as optional input. If not provided, it will read the network configuration
    from the network.ini in the output folder.
    """

    _parser: ConfigParser

    def __init__(self) -> None:
        self._parser = ConfigParser(
            inline_comment_prefixes="#",
            converters={"list": lambda x: [x.strip() for x in x.split(",")]},
        )

    def read(self, ini_file: Path) -> AnalysisConfigData:
        if not isinstance(ini_file, Path) or not ini_file.is_file():
            raise ValueError("No analysis ini file 'Path' provided.")
        self._parser.read(ini_file)
        self._remove_none_values()

        _parent_dir = ini_file.parent
        _config_data = AnalysisConfigData(
            root_path=_parent_dir.parent,
            input_path=_parent_dir.joinpath("input"),
            static_path=_parent_dir.joinpath("static"),
            output_path=_parent_dir.joinpath("output"),
            **self._get_sections(),
        )
        _config_data.project.name = _parent_dir.name

        self._copy_output_files(ini_file, _config_data)

        return _config_data

    def _remove_none_values(self) -> None:
        # Remove 'None' from values, replace them with empty strings
        for _section in self._parser.sections():
            _keys_with_none = [
                k for k, v in self._parser[_section].items() if v == "None"
            ]
            for _key_with_none in _keys_with_none:
                self._parser[_section].pop(_key_with_none)

    def _get_sections(self) -> dict:
        return {
            "project": self.get_project_section(),
            "analyses": self.get_analysis_sections(),
        }

    def get_project_section(self) -> ProjectSection:
        return ProjectSection(**self._parser["project"])

    def _get_analysis_section_losses(self, section_name: str) -> AnalysisSectionLosses:
        _section = AnalysisSectionLosses(**self._parser[section_name])
        _section.analysis = AnalysisLossesEnum.get_enum(
            self._parser.get(section_name, "analysis", fallback=None)
        )
        _section.save_gpkg = self._parser.getboolean(
            section_name, "save_gpkg", fallback=_section.save_gpkg
        )
        _section.save_csv = self._parser.getboolean(
            section_name, "save_csv", fallback=_section.save_csv
        )
        _weighing = self._parser.get(section_name, "weighing", fallback=None)
        # Map distance -> length
        if _weighing == "distance":
            _section.weighing = WeighingEnum.LENGTH
        else:
            _section.weighing = WeighingEnum.get_enum(_weighing)
        _section.loss_type = LossTypeEnum.get_enum(
            self._parser.get(section_name, "loss_type", fallback=None)
        )
        # losses
        _section.traffic_cols = self._parser.getlist(
            section_name, "traffic_cols", fallback=_section.traffic_cols
        )
        _section.traffic_period = TrafficPeriodEnum.get_enum(
            self._parser.get(section_name, "traffic_period", fallback=None)
        )
        _section.hours_per_traffic_period = self._parser.getint(
            section_name,
            "hours_per_traffic_period",
            fallback=_section.hours_per_traffic_period,
        )
        _section.trip_purposes = list(
            map(
                TripPurposeEnum.get_enum,
                self._parser.getlist(section_name, "trip_purposes", fallback=[]),
            )
        )
        _section.production_loss_per_capita_per_hour = self._parser.getfloat(
            section_name,
            "production_loss_per_capita_per_hour",
            fallback=_section.production_loss_per_capita_per_hour,
        )

        # accessibility analyses
        _section.threshold = self._parser.getfloat(
            section_name,
            "threshold",
            fallback=_section.threshold,
        )
        _section.threshold_destinations = self._parser.getfloat(
            section_name,
            "threshold_destinations",
            fallback=_section.threshold_destinations,
        )
        _section.uniform_duration = self._parser.getfloat(
            section_name,
            "uniform_duration",
            fallback=_section.uniform_duration,
        )
        _section.gdp_percapita = self._parser.getfloat(
            section_name,
            "gdp_percapita",
            fallback=_section.gdp_percapita,
        )
        _section.calculate_route_without_disruption = self._parser.getboolean(
            section_name,
            "calculate_route_without_disruption",
            fallback=_section.calculate_route_without_disruption,
        )
        _section.buffer_meters = self._parser.getfloat(
            section_name,
            "buffer_meters",
            fallback=_section.buffer_meters,
        )
        _section.threshold_locations = self._parser.getfloat(
            section_name,
            "threshold_locations",
            fallback=_section.threshold_locations,
        )
        _section.save_traffic = self._parser.getboolean(
            section_name, "save_traffic", fallback=_section.save_traffic
        )

        return _section

    def _get_analysis_section_damages(
        self, section_name: str
    ) -> AnalysisSectionDamages:
        _section = AnalysisSectionDamages(**self._parser[section_name])
        _section.analysis = AnalysisDamagesEnum.get_enum(
            self._parser.get(section_name, "analysis", fallback=None)
        )
        _section.save_gpkg = self._parser.getboolean(
            section_name, "save_gpkg", fallback=_section.save_gpkg
        )
        _section.save_csv = self._parser.getboolean(
            section_name, "save_csv", fallback=_section.save_csv
        )
        # adaptation/effectiveness measures
        _section.return_period = self._parser.getfloat(
            section_name,
            "return_period",
            fallback=_section.return_period,
        )
        _section.repair_costs = self._parser.getfloat(
            section_name,
            "repair_costs",
            fallback=_section.repair_costs,
        )
        _section.evaluation_period = self._parser.getfloat(
            section_name,
            "evaluation_period",
            fallback=_section.evaluation_period,
        )
        _section.evaluation_period = self._parser.getfloat(
            section_name,
            "representative_damage_percentage",
            fallback=_section.representative_damage_percentage,
        )
        _section.interest_rate = self._parser.getfloat(
            section_name,
            "interest_rate",
            fallback=_section.interest_rate,
        )
        _section.climate_factor = self._parser.getfloat(
            section_name,
            "climate_factor",
            fallback=_section.climate_factor,
        )
        _section.climate_period = self._parser.getfloat(
            section_name,
            "climate_period",
            fallback=_section.climate_period,
        )
        # road damage
        _section.event_type = EventTypeEnum.get_enum(
            self._parser.get(section_name, "event_type", fallback=None)
        )
        _section.risk_calculation_mode = RiskCalculationModeEnum.get_enum(
            self._parser.get(section_name, "risk_calculation_mode", fallback=None)
        )
        _section.risk_calculation_year = self._parser.getint(
            section_name,
            "risk_calculation_year",
            fallback=_section.risk_calculation_year,
        )
        _section.damage_curve = DamageCurveEnum.get_enum(
            self._parser.get(section_name, "damage_curve", fallback=None)
        )
        _section.create_table = self._parser.getboolean(
            section_name,
            "create_table",
            fallback=_section.create_table,
        )
        return _section

    def get_analysis_sections(self) -> list[AnalysisSectionBase]:
        """
        Extracts info from [analysis<n>] sections

        Returns:
            list[AnalysisSection]: List of analyses (both damages and losses)
        """
        _analysis_sections = []

        _section_names = list(
            section_name
            for section_name in self._parser.sections()
            if "analysis" in section_name
        )
        for _section_name in _section_names:
            _analysis_type = self._parser.get(_section_name, "analysis")
            if _analysis_type in DamagesAnalysisNameList:
                _analysis_section = self._get_analysis_section_damages(_section_name)
            elif _analysis_type in LossesAnalysisNameList:
                _analysis_section = self._get_analysis_section_losses(_section_name)
            else:
                raise ValueError(f"Analysis {_analysis_type} not supported.")
            _analysis_sections.append(_analysis_section)

        return _analysis_sections

    def _copy_output_files(
        self, from_path: Path, config_data: AnalysisConfigData
    ) -> None:
        _output_dir = config_data.root_path.joinpath(config_data.project.name, "output")
        config_data.output_path = _output_dir
        if not _output_dir.exists():
            _output_dir.mkdir(parents=True)
        try:
            copyfile(
                from_path,
                config_data.output_path.joinpath("{}.ini".format(from_path.stem)),
            )
        except FileNotFoundError as e:
            logging.warning(e)

    def _parse_path_list(
        self, property_name: str, path_list: str, config_data: AnalysisConfigData
    ) -> list[Path]:
        _list_paths = []
        for path_value in path_list.split(","):
            path_value = Path(path_value)
            if path_value.is_file():
                _list_paths.append(path_value)
                continue

            _project_name_dir = config_data.root_path.joinpath(config_data.project.name)
            abs_path = _project_name_dir.joinpath("static", property_name, path_value)
            try:
                assert abs_path.is_file()
            except AssertionError:
                abs_path = _project_name_dir.joinpath(
                    "input", property_name, path_value
                )

            _list_paths.append(abs_path)
        return _list_paths



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable

from geopandas import GeoDataFrame
from networkx import Graph, set_edge_attributes
from numpy import nan
from rasterstats import zonal_stats
from tqdm import tqdm

from ra2ce.network.hazard.hazard_common_functions import (
    get_edges_geoms,
    validate_extent_graph,
)
from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_base import (
    HazardIntersectBuilderBase,
)
from ra2ce.network.networks_utils import (
    fraction_flooded,
    get_graph_edges_extent,
    get_valid_mean,
)


@dataclass
class HazardIntersectBuilderForTif(HazardIntersectBuilderBase):
    hazard_aggregate_wl: str = ""
    hazard_names: list[str] = field(default_factory=list)
    ra2ce_names: list[str] = field(default_factory=list)
    hazard_tif_files: list[Path] = field(default_factory=list)

    @property
    def _combined_names(self) -> list[tuple[str, str]]:
        """
        Combines into tuples `hazard_names` and `ra2ce_names`.

        Returns:
            list[tuple[str, str]]: List of `str` tuples (hazard_name, ra2ce_name).
        """
        return list(zip(self.hazard_names, self.ra2ce_names))

    def _from_networkx(self, hazard_overlay: Graph) -> Graph:
        """Overlays the hazard raster over the road segments graph.

        Args:
            *hf* (list of Pathlib paths) : #not sure if this is needed as argument if we also read if from the config
            *graph* (NetworkX Graph) : NetworkX graph with geometries that will be intersected with the hazard map raster.

        Returns:
            *graph* (NetworkX Graph) : NetworkX graph with hazard values
        """
        # TODO apply multiprocessing?
        # from tqdm import (
        #     tqdm,  # somehow this only works when importing here and not at the top of the file
        # )

        # Verify the graph type (networkx)
        assert isinstance(hazard_overlay, Graph)
        extent_graph = get_graph_edges_extent(hazard_overlay)

        # Get all edge geometries
        edges_geoms = get_edges_geoms(hazard_overlay)

        def overlay_network_x(hazard_tif_file: Path, hazard_name: str, ra2ce_name: str):
            # Check if the hazard and graph extents overlap
            validate_extent_graph(extent_graph, hazard_tif_file)
            # Add a no-data value for the edges that do not have a geometry
            set_edge_attributes(
                hazard_overlay,
                {
                    (u, v, k): {ra2ce_name + "_" + self.hazard_aggregate_wl[:2]: nan}
                    for u, v, k, edata in hazard_overlay.edges.data(keys=True)
                    if "geometry" not in edata
                },
            )

            # Add the hazard values to the edges that do have a geometry
            gdf = GeoDataFrame(
                {"geometry": [edata["geometry"] for u, v, k, edata in edges_geoms]}
            )
            tqdm.pandas(desc="Graph hazard overlay with " + hazard_name)
            _tif_hazard_files = str(hazard_tif_file)
            if self.hazard_aggregate_wl == "mean":
                flood_stats = gdf.geometry.progress_apply(
                    lambda x, _files_value=_tif_hazard_files: zonal_stats(
                        x,
                        _files_value,
                        all_touched=True,
                        add_stats={"mean": get_valid_mean},
                    )
                )
            else:
                flood_stats = gdf.geometry.progress_apply(
                    lambda x, _files_value=_tif_hazard_files: zonal_stats(
                        x,
                        _files_value,
                        all_touched=True,
                        stats=f"{self.hazard_aggregate_wl}",
                    )
                )

            try:
                flood_stats = flood_stats.apply(
                    lambda x: (
                        x[0][self.hazard_aggregate_wl]
                        if x[0][self.hazard_aggregate_wl]
                        else 0
                    )
                )
                set_edge_attributes(
                    hazard_overlay,
                    {
                        (edges[0], edges[1], edges[2]): {
                            ra2ce_name + "_" + self.hazard_aggregate_wl[:2]: x
                        }
                        for x, edges in zip(flood_stats, edges_geoms)
                    },
                )
            except KeyError:
                logging.warning(
                    "No aggregation method ('aggregate_wl') is chosen - choose from 'max', 'min' or 'mean'."
                )

            # Get the fraction of the road that is intersecting with the hazard
            tqdm.pandas(desc="Graph fraction with hazard overlay with " + hazard_name)
            graph_fraction_flooded = gdf.geometry.progress_apply(
                lambda x, _files_values=_tif_hazard_files: fraction_flooded(
                    x, _files_values
                )
            )
            graph_fraction_flooded = graph_fraction_flooded.fillna(0)
            set_edge_attributes(
                hazard_overlay,
                {
                    (edges[0], edges[1], edges[2]): {ra2ce_name + "_fr": x}
                    for x, edges in zip(graph_fraction_flooded, edges_geoms)
                },
            )

        self._overlay_hazard_files(overlay_network_x)
        return hazard_overlay

    def _from_geodataframe(self, hazard_overlay: GeoDataFrame):
        """Overlays the hazard raster over the road segments GeoDataFrame.

        Args:
            *graph* (GeoDataFrame) : GeoDataFrame that will be intersected with the hazard map raster.

        Returns:

        """
        assert isinstance(hazard_overlay, GeoDataFrame), "Network is not a GeoDataFrame"

        # Make sure none of the geometries is a nonetype object (this will raise an error in zonal_stats)
        empty_entries = hazard_overlay.loc[hazard_overlay.geometry.isnull()]
        if any(empty_entries):
            logging.warning(
                (
                    "Some geometries have NoneType objects (no coordinate information), namely: {}.".format(
                        empty_entries
                    )
                    + "This could be due to segmentation, and might cause an exception in hazard overlay"
                )
            )

        def overlay_geodataframe(
            hazard_tif_file: Path, hazard_name: str, ra2ce_name: str
        ):
            # Validate input
            # Check if network and raster overlap
            extent_graph = (
                hazard_overlay.total_bounds[0],
                hazard_overlay.total_bounds[2],
                hazard_overlay.total_bounds[1],
                hazard_overlay.total_bounds[3],
            )
            validate_extent_graph(extent_graph, hazard_tif_file)

            tqdm.pandas(desc="Network hazard overlay with " + hazard_name)
            _hazard_files_str = str(hazard_tif_file)
            # Performance sinkhole
            flood_stats = hazard_overlay.geometry.progress_apply(
                lambda _geom_vector: zonal_stats(
                    vectors=_geom_vector,
                    raster=_hazard_files_str,
                    all_touched=True,
                    stats="min max",
                    add_stats={"mean": get_valid_mean},
                )
            )

            def _get_attributes(gen_flood_stat: list[dict]) -> tuple:
                # Just get the first element of the generator
                _flood_stat = gen_flood_stat[0]
                return _flood_stat["min"], _flood_stat["max"], _flood_stat["mean"]

            (
                hazard_overlay[ra2ce_name + "_mi"],
                hazard_overlay[ra2ce_name + "_ma"],
                hazard_overlay[ra2ce_name + "_me"],
            ) = list(zip(*map(_get_attributes, flood_stats)))

            tqdm.pandas(desc="Network fraction with hazard overlay with " + hazard_name)
            hazard_overlay[ra2ce_name + "_fr"] = hazard_overlay.geometry.progress_apply(
                lambda x, _hz_str=_hazard_files_str: fraction_flooded(x, _hz_str)
            )

        self._overlay_hazard_files(overlay_geodataframe)
        return hazard_overlay

    def _overlay_hazard_files(self, overlay_func: Callable[[str, str, str], None]):
        for i, (hn, rn) in enumerate(self._combined_names):
            overlay_func(self.hazard_tif_files[i], hn, rn)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-
import copy
import logging
from collections import defaultdict
from typing import Any

import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely.geometry import LineString, MultiLineString
from tqdm import tqdm

from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper


class OriginClosestDestination:
    """The origin closest destination analyses using NetworkX graphs.

    Attributes:
        file_id: The file id of the network.
        origins_destinations: An object with the origins and destinations.
        analysis: An object with the analysis details.
        graph_file: An object with the NetworkX graph.
        graph_file_hazard: An object with the NetworkX graph with hazard disruption.
        hazard_names: An object with the hazard names.
    """

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,  # noqa: F821
    ):
        self.analysis = analysis_input.analysis
        self.graph_file = analysis_input.graph_file
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.static_path = analysis_input.static_path
        self.hazard_names = analysis_input.hazard_names

        self.crs = 4326  # TODO PUT IN DOCUMENTATION OR MAKE CHANGEABLE
        self.unit = "km"
        self.network_threshold = self.analysis.threshold
        self.threshold_destinations = self.analysis.threshold_destinations
        self.weighing = self.analysis.weighing.config_value
        self.origins_destinations = analysis_input.origins_destinations
        self.o_name = self.origins_destinations.origins_names
        self.d_name = self.origins_destinations.destinations_names
        self.od_id = self.origins_destinations.id_name_origin_destination
        self.origin_out_fraction = self.origins_destinations.origin_out_fraction
        self.origin_count = self.origins_destinations.origin_count
        self.od_key = "od_id"
        self.id_name = (
            analysis_input.file_id if analysis_input.file_id is not None else "rfid"
        )

        self.destination_names = None
        self.destination_key = None
        if self.origins_destinations.category:
            self.destination_key = "category"
            self.destination_key_value = self.origins_destinations.category

        self.results_dict = {}

    def optimal_route_origin_closest_destination(
        self,
    ) -> tuple[nx.MultiGraph, gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """Calculates per origin the location of its closest destination"""
        graph = self.graph_file.get_graph()

        # Load the origins and destinations
        origins = self.load_origins()
        destinations = self.load_destinations()

        # Create a copy of the graph to save all the results in
        base_graph = copy.deepcopy(graph)

        # Add a column for the number of people that go to a certain destination, per flood map
        col_name = "noHaz"

        if self.destination_key:
            self.destination_names = list(
                destinations[self.origins_destinations.category].unique()
            )
            self.destination_names_short = {
                dn: f"D{i+1}" for i, dn in enumerate(self.destination_names)
            }
            logging.info(self.destination_names_short)  # TODO: WRITE SOMEWHERE

            for i, dn in self.destination_names_short.items():
                destinations[col_name + "_P" + dn] = 0

            (
                base_graph,
                origins,
                destinations,
                list_disrupted_dest,
                optimal_routes_gdf,
            ) = self.find_multiple_closest_locations(
                graph, base_graph, origins, destinations, col_name
            )
        else:
            destinations[col_name + "_P"] = 0
            (
                base_graph,
                origins,
                destinations,
                list_disrupted_dest,
                optimal_routes_gdf,
            ) = self.find_closest_location(
                graph, base_graph, origins, destinations, col_name
            )

        cnt_per_destination = (
            optimal_routes_gdf.groupby("destination")["origin_cnt"].sum().reset_index()
        )
        for dest, origin_cnt in zip(
            cnt_per_destination["destination"],
            cnt_per_destination["origin_cnt"],
        ):
            destinations.loc[
                destinations[self.od_id] == int(dest.split("_")[-1]), "origin_cnt"
            ] = origin_cnt

        return base_graph, optimal_routes_gdf, destinations

    def multi_link_origin_closest_destination(
        self,
    ) -> tuple[
        nx.MultiGraph,
        gpd.GeoDataFrame,
        gpd.GeoDataFrame,
        pd.DataFrame,
        gpd.GeoDataFrame,
    ]:
        """Calculates per origin the location of its closest destination with hazard disruption"""
        graph = self.graph_file_hazard.get_graph()

        # Load the origins and destinations
        origins = self.load_origins()
        destinations = self.load_destinations()

        # Create a copy of the graph to save all the results in
        base_graph = copy.deepcopy(graph)

        if self.destination_key:
            self.destination_names = list(
                destinations[self.origins_destinations.category].unique()
            )
            self.destination_names_short = {
                dn: f"D{i+1}" for i, dn in enumerate(self.destination_names)
            }
            logging.info(self.destination_names_short)  # TODO: WRITE SOMEWHERE

        aggregated = []
        opt_routes_aggregated = []

        # Calculate the criticality
        hazards = [
            self.hazard_names.names_df.loc[
                self.hazard_names.names_df["File name"] == hazard, "RA2CE name"
            ].values[0]
            for hazard in self.hazard_names.names
        ]
        hazards.sort()
        for hazard_name in hazards:
            self.results_dict = dict()
            self.results_dict["Hazard"] = [hazard_name]

            # Add a column for the number of people that go to a certain destination, per flood map
            if self.destination_key:
                for i, dn in self.destination_names_short.items():
                    destinations[hazard_name + "_P" + dn] = 0
            else:
                destinations[hazard_name + "_P"] = 0

            # Check if the o/d pairs are still connected while some links are disrupted by the hazard(s)
            h = copy.deepcopy(graph)
            edges_remove = []
            for e in graph.edges.data(keys=True):
                if (hazard_name in e[-1]) and (
                    ("bridge" not in e[-1])
                    or ("bridge" in e[-1] and e[-1]["bridge"] != "yes")
                ):
                    edges_remove.append(e)
            edges_remove = [e for e in edges_remove if (e[-1][hazard_name] is not None)]
            edges_remove = [
                e
                for e in edges_remove
                if (e[-1][hazard_name] > float(self.network_threshold))
            ]
            h.remove_edges_from(edges_remove)

            if self.destination_key:
                (
                    base_graph,
                    origins,
                    destinations,
                    list_disrupted_dest,
                    optimal_routes_gdf,
                ) = self.find_multiple_closest_locations(
                    h,
                    base_graph,
                    origins,
                    destinations,
                    hazard_name,
                    hazard_name,
                )
            else:
                (
                    base_graph,
                    origins,
                    destinations,
                    list_disrupted_dest,
                    optimal_routes_gdf,
                ) = self.find_closest_location(
                    h,
                    base_graph,
                    origins,
                    destinations,
                    hazard_name,
                    hazard_name,
                )

            ## THE BELOW IS NOT USED AT THE MOMENT - COULD BE IMPLEMENTED AS OPTION AT A LATER STAGE ##
            # # Now calculate for the routes that were going to a flooded destination, another non-flooded destination
            # # There can be nodes that are both origins and destinations. The flooded destinations are removed from
            # # the graph.
            # list_dests_flooded_destination_only = [dest[-1][0] for dest in list_disrupted_dest if
            #                                        not self.o_name in dest[-1][-1]]
            # list_dests_flooded_destination_and_origin = [dest[-1][0] for dest in list_disrupted_dest if
            #     self.o_name in dest[-1][-1] and self.d_name in dest[-1][-1]]
            # list_origins_without_routes = [dest[0][0] for dest in list_disrupted_dest]
            # list_origins_with_routes = [n for n in h.nodes() if n not in list_origins_without_routes]
            #
            # disr_by_flood = len(list_dests_flooded_destination_only + list_dests_flooded_destination_and_origin)
            #
            # if disr_by_flood > 0:
            #     # Remove the flooded destination nodes
            #     h.remove_nodes_from(list_dests_flooded_destination_only)
            #
            #     for node in list_dests_flooded_destination_and_origin:
            #         # Delete the destination name from the "od_id" attribute of the nodes of which the destination is
            #         # flooded.
            #         od_id = h.nodes[node]["od_id"]
            #         h.nodes[node]["od_id"] = ",".join([od for od in od_id.split(",") if self.d_name not in od])
            #
            #     # Remove the origin nodes that already have a destination
            #     h.remove_nodes_from(list_origins_with_routes)
            #
            #     # Search for the next closest location
            #     base_graph, origins, destinations, list_disrupted_dest, results_dict_update = self.find_closest_location(h,
            #                                                                                                        base_graph,
            #                                                                                                        origins,
            #                                                                                                        destinations,
            #                                                                                                        hazard_name,
            #                                                                                                        self.d_name)
            #
            #     self.results_dict['Nr. no access'] = [
            #         self.results_dict['Nr. no access'][0] + self.results_dict_update['Nr. no access'][0]]
            #
            # self.results_dict.update({
            #     "Flooded destination": [disr_by_flood]
            # })
            ## THE ABOVE IS NOT USED AT THE MOMENT - COULD BE IMPLEMENTED AS OPTION AT A LATER STAGE ##

            aggregated.append(pd.DataFrame(self.results_dict))

            optimal_routes_gdf["name"] = hazard_name
            opt_routes_aggregated.append(optimal_routes_gdf)

        # Create one dataframe from the aggregated results
        aggregated = pd.concat(aggregated)
        opt_routes_aggregated = pd.concat(opt_routes_aggregated)

        return base_graph, origins, destinations, aggregated, opt_routes_aggregated

    def get_route_length(
        self, graph: nx.MultiGraph, origin_node: int, destination_node: int
    ) -> float:
        # calculate the length of the preferred route
        return nx.dijkstra_path_length(
            graph, origin_node, destination_node, weight=self.weighing
        )

    def compare_route_with_without_disruption(
        self,
        pref_routes: gpd.GeoDataFrame,
        nr_per_route: float,
        o_name: str,
        d_name: str,
        alt_route: float,
        alt_dist: float,
    ) -> None:
        pp_no_delay = [0]
        pp_delayed = [0]
        extra_weights = [0]
        extra_kms_total = [0]

        # If the destination is different from the origin, the destination is further than without hazard disruption
        if pref_routes.loc[
            (pref_routes["origin"] == o_name) & (pref_routes["destination"] == d_name)
        ].empty:
            # subtract the length/time of the optimal route from the alternative route
            extra_dist = (
                alt_route
                - pref_routes.loc[pref_routes["origin"] == o_name, self.weighing].iloc[
                    0
                ]
            )
            extra_km = (
                alt_dist
                - pref_routes.loc[pref_routes["origin"] == o_name, "tot_km"].iloc[0]
            )
            pp_delayed.append(nr_per_route)
            extra_weights.append(extra_dist)
            extra_kms_total.append(extra_km)
        else:
            pp_no_delay.append(nr_per_route)

        if pp_no_delay != [0]:
            self.results_dict["Nr. no delay"] = [round(sum(pp_no_delay))]
        if pp_delayed != [0]:
            self.results_dict["Nr. delayed"] = [round(sum(pp_delayed))]
        if extra_weights != [0]:
            self.results_dict[f"Total extra detour {self.weighing}"] = [
                sum(extra_weights)
            ]
        if extra_kms_total != [0]:
            self.results_dict[f"Total extra detour distance ({self.unit})"] = [
                sum(extra_kms_total)
            ]

    def get_route_path(
        self,
        graph: nx.MultiGraph,
        base_graph: nx.MultiGraph,
        origin_node: int,
        destination_node: int,
        nr_from_origin: float,
        col_name: str,
    ) -> tuple[nx.MultiGraph, float, MultiLineString]:
        # Get the nodes of the optimal route
        route_nodes = nx.dijkstra_path(
            graph, origin_node, destination_node, weight=self.weighing
        )

        # find out which edges belong to the preferred path
        edgesinpath = list(zip(route_nodes[0:], route_nodes[1:]))

        # calculate the total length of the alternative route
        # Find the road segments that are used for the detour to the same or another destination
        pref_edges = []
        length_list = []
        for u, v in edgesinpath:
            # get edge with the lowest weighing if there are multiple edges that connect u and v
            _uv_graph = graph[u][v]
            edge_key = sorted(
                _uv_graph, key=lambda x, _fgraph=_uv_graph: _fgraph[x][self.weighing]
            )[0]
            _uv_graph_edge = _uv_graph[edge_key]
            if "geometry" in _uv_graph_edge:
                pref_edges.append(_uv_graph_edge["geometry"])
            else:
                pref_edges.append(
                    LineString([graph.nodes[u]["geometry"], graph.nodes[v]["geometry"]])
                )

            if "length" in _uv_graph_edge:
                length_list.append(_uv_graph_edge["length"])

            # Add the number of people that need to go to a destination to the road segments. For now, each road segment in a route
            # gets attributed all the people that are taking that route.
            base_graph[u][v][edge_key][col_name] = (
                base_graph[u][v][edge_key][col_name] + nr_from_origin
            )

        pref_edges = MultiLineString(pref_edges)

        return base_graph, sum(length_list), pref_edges

    def get_nr_people_on_route(
        self, origins: gpd.GeoDataFrame, origin_node: str
    ) -> float:
        # Find the number of people per neighborhood
        try:
            nr_people_per_route_total = origins.loc[
                origins[self.od_id] == int(origin_node.split("_")[-1]),
                self.origin_count,
            ].iloc[0]
        except IndexError:
            origin_node = [a for a in origin_node.split(",") if self.o_name in a][0]
            nr_people_per_route_total = origins.loc[
                origins[self.od_id] == int(origin_node.split("_")[-1]),
                self.origin_count,
            ].iloc[0]
        nr_per_route = nr_people_per_route_total * self.origin_out_fraction
        return nr_per_route

    def update_destinations(
        self,
        destinations: gpd.GeoDataFrame,
        destination_name: str,
        nr_per_route: float,
        col_name: str,
    ) -> gpd.GeoDataFrame:
        dest_ids = [
            int(d.split("_")[-1])
            for d in [dest for dest in destination_name.split(",")]
            if self.d_name in d
        ]

        # Add the number of people to the total number of people that go to that destination
        destinations.loc[destinations[self.od_id].isin(dest_ids), col_name] = (
            destinations.loc[destinations[self.od_id].isin(dest_ids), col_name].sum()
            + nr_per_route
        )
        return destinations

    def update_origins(
        self, origins: gpd.GeoDataFrame, other: list, col_name: str
    ) -> gpd.GeoDataFrame:
        # Attribute to the origins that don't have access that they do not have any access
        if len(other) > 0:
            for oth in other:
                od_id_list = oth[-1].split(",")
                for od_id in od_id_list:
                    origins.loc[
                        origins[self.od_id] == int(od_id.split("_")[-1]),
                        col_name,
                    ] = "no access"
        return origins

    def get_nr_without_access(
        self,
        origins: gpd.GeoDataFrame,
        origins_without_access: list,
        add_key_name: str = None,
    ) -> None:
        # Calculate the number of people that cannot access any destination
        origins_no_access = [
            o
            for o in [
                oth[-1].split(",") if len(origins_without_access) > 0 else 0
                for oth in origins_without_access
            ]
        ]
        origins_no_access = [
            int(item.split("_")[-1])
            for sublist in origins_no_access
            for item in sublist
        ]

        pp_no_access = [
            (
                round(
                    (
                        origins.loc[
                            origins[self.od_id].isin(origins_no_access),
                            self.origin_count,
                        ]
                        * self.origin_out_fraction
                    ).sum()
                )
                if len(origins_no_access) > 0
                else 0
            )
        ]

        self.results_dict[f"Nr. no access{add_key_name}"] = pp_no_access

    def find_closest_location(
        self,
        disrupted_graph: nx.MultiGraph,
        base_graph: nx.MultiGraph,
        origins: gpd.GeoDataFrame,
        destinations: gpd.GeoDataFrame,
        column_name: str,
        hazard_name: str = None,
        pref_routes: gpd.GeoDataFrame = None,
    ) -> tuple[
        nx.MultiGraph,
        gpd.GeoDataFrame,
        gpd.GeoDataFrame,
        list,
        gpd.GeoDataFrame,
    ]:
        """Find the closest destination nodes with a certain attribute from all origin nodes for a single destination
        Args:
            disrupted_graph: NetworkX graph with origin and destination nodes and disrupted edges removed
            base_graph: Original NetworkX graph
            origins: GeoDataFrame containing the origins
            destinations: GeoDataFrame containing the destinations
            column_name: the name of the column which will be added to the base_graph, origins, destinations, and self.results_dict
            hazard_name: the name of the hazard (which is an attribute of the edges in the disrupted_graph)
            pref_routes: GeoDataFrame containing the preferred routes without disruption

        Returns:
            base_graph:
            origins: the updated origins GeoDataFrame
            destinations: the updated destinations GeoDataFrame
            list_disrupted_destinations [list of tuples]: list of the origin and destination node id and node name from the origins/nodes that do not have a route between them
        """
        name_save = column_name + "_{}"
        nx.set_edge_attributes(base_graph, 0, name_save.format("P"))

        # Add a column to the neighborhoods, to indicate if they have access to any destination POI.
        # The origins without access are indicated later
        origins[name_save.format("A")] = "access"

        disrupted_graph.add_node("special", speciallabel="special")

        special_edges = []
        for n, ndat in disrupted_graph.nodes.data():
            if self.od_key in ndat and self.d_name in ndat[self.od_key]:
                special_edges.append((n, "special", {self.weighing: 0}))

        disrupted_graph.add_edges_from(special_edges)

        optimal_routes = []
        list_disrupted_destinations = []
        list_no_path = []
        node_checked_has_path = dict()
        for n_ndat in tqdm(disrupted_graph.nodes.data(), desc="Finding optimal routes"):
            _new_origins = self._find_optimal_routes(
                node_checked_has_path,
                list_no_path,
                n_ndat,
                disrupted_graph,
                hazard_name,
                list_disrupted_destinations,
                pref_routes,
                "special",
                name_save,
                optimal_routes,
                origins,
                base_graph,
                destinations,
            )

            if _new_origins is not None:
                origins = _new_origins
                self.get_nr_without_access(origins, list_no_path)

        # Remove the special edges
        disrupted_graph.remove_edges_from(
            [(n[0], n[1]) for n in disrupted_graph.edges.data() if n[1] == "special"]
        )

        # Remove the closest attribute from the nodes
        origin_closest_dst = [
            nn[0] for nn in disrupted_graph.nodes.data() if "closest" in nn[-1]
        ]

        if origin_closest_dst:
            for o in origin_closest_dst:
                del disrupted_graph.nodes[o]["closest"]

            optimal_routes_gdf = pd.concat(optimal_routes)

        return (
            base_graph,
            origins,
            destinations,
            list_disrupted_destinations,
            optimal_routes_gdf,
        )

    def _find_optimal_routes(
        self,
        node_checked_has_path: dict,
        list_no_path: list,
        n_ndat: tuple[int, dict[str, Any]],
        disrupted_graph: nx.MultiGraph,
        hazard_name: str,
        list_disrupted_destinations: list,
        pref_routes: gpd.GeoDataFrame,
        dest_name: str,
        name_save: str,
        optimal_routes: list,
        origins: gpd.GeoDataFrame,
        base_graph: nx.MultiGraph,
        destinations: gpd.GeoDataFrame,
    ) -> gpd.GeoDataFrame:
        """
        Refactored method to avoid duplication of code between `find_closest_location` and `find_multiple_closest_locations` with subtile differences:
        - The first would not use a `dest_name` attribute.
        - The second one would use 'ndat["closest"]' instead of the assigned 'closest_dest'

        Returns:
            Optional[gpd.GeoDataFrame]: When the wrapper for-loop needs to go into the next iteration it will return 'None'. Otherwise a resulting `gpd.GeoDataFrame`.
        """
        n, ndat = n_ndat
        if self.od_key in ndat and self.o_name in ndat[self.od_key]:
            if nx.has_path(disrupted_graph, n, dest_name):
                # Add elements to the dictionary this way to prevent an exception when
                # their key is not present.
                node_checked_has_path.setdefault(ndat[self.od_key], []).append(n)
                path = nx.shortest_path(
                    disrupted_graph,
                    source=n,
                    target=dest_name,
                    weight=self.weighing,
                )
                # Closest node with destLabelContains in keyName
                ndat["closest"] = path[-2]
                closest_dest = ndat["closest"]

                # Check if the destination that is accessed, is flooded
                if hazard_name:
                    try:
                        if (
                            disrupted_graph.nodes[closest_dest][hazard_name]
                            > self.threshold_destinations
                        ):
                            list_disrupted_destinations.append(
                                (
                                    (n, ndat[self.od_key]),
                                    (
                                        closest_dest,
                                        disrupted_graph.nodes[closest_dest][
                                            self.od_key
                                        ],
                                    ),
                                )
                            )
                            return None
                    except KeyError as e:
                        logging.error(
                            f"The destination nodes do not contain the required attribute '{hazard_name}',"
                            f" please make sure that the hazard overlay is done correctly by rerunning the 'network.ini'"
                            f" and checking the output files."
                        )
                        raise e

                nr_per_route = self.get_nr_people_on_route(origins, ndat[self.od_key])
                base_graph, route_path, route_geoms = self.get_route_path(
                    disrupted_graph,
                    base_graph,
                    n,
                    closest_dest,
                    nr_per_route,
                    name_save.format("P"),
                )
                if pref_routes:
                    route_length = self.get_route_length(
                        disrupted_graph, n, closest_dest
                    )
                    self.compare_route_with_without_disruption(
                        pref_routes,
                        nr_per_route,
                        ndat[self.od_key],
                        disrupted_graph.nodes[closest_dest],
                        route_length,
                        route_path,
                    )
                destinations = self.update_destinations(
                    destinations,
                    disrupted_graph.nodes[closest_dest][self.od_key],
                    nr_per_route,
                    name_save.format("P"),
                )

                if route_geoms:
                    optimal_routes.append(
                        gpd.GeoDataFrame(
                            {
                                "o_node": n,
                                "d_node": closest_dest,
                                "origin": ndat[self.od_key],
                                "destination": disrupted_graph.nodes[closest_dest][
                                    self.od_key
                                ],
                                self.weighing: route_path,
                                "origin_cnt": nr_per_route,
                                "geometry": [route_geoms],
                                "category": dest_name,
                            },
                            crs=self.crs,
                        )
                    )
            else:
                if ndat[self.od_key] not in node_checked_has_path:
                    list_no_path.append((n, ndat[self.od_key]))

        return self.update_origins(origins, list_no_path, name_save.format("A"))

    def find_multiple_closest_locations(
        self,
        disrupted_graph: nx.MultiGraph,
        base_graph: nx.MultiGraph,
        origins: gpd.GeoDataFrame,
        destinations: gpd.GeoDataFrame,
        column_name: str,
        hazard_name: str = None,
        pref_routes: gpd.GeoDataFrame = None,
    ) -> tuple[
        nx.MultiGraph,
        gpd.GeoDataFrame,
        gpd.GeoDataFrame,
        list,
        gpd.GeoDataFrame,
    ]:
        """Find the closest destination nodes with a certain attribute from all origin nodes for multiple destinations
        Args:
            disrupted_graph: NetworkX graph with origin and destination nodes and disrupted edges removed
            base_graph: Original NetworkX graph
            origins: GeoDataFrame containing the origins
            destinations: GeoDataFrame containing the destinations
            column_name: the name of the column which will be added to the base_graph, origins, destinations, and self.results_dict
            hazard_name: the name of the hazard
            pref_routes: GeoDataFrame containing the preferred routes without disruption

        Returns:
            base_graph:
            origins: the updated origins GeoDataFrame
            destinations: the updated destinations GeoDataFrame
            list_disrupted_destinations [list of tuples]: list of the origin and destination node id and node name from the origins/nodes that do not have a route between them
        """

        optimal_routes = []
        for dest_name in self.destination_names:
            name_save = column_name + "_{}" + self.destination_names_short[dest_name]
            nx.set_edge_attributes(base_graph, 0, name_save.format("P"))

            # Add a column to the neighborhoods, to indicate if they have access to any destination POI.
            # The origins without access are indicated later
            origins[name_save.format("A")] = "access"

            disrupted_graph.add_node(dest_name, speciallabel=dest_name)

            special_edges = []
            for n, ndat in disrupted_graph.nodes.data():
                if (
                    self.od_key in ndat
                    and self.destination_key in ndat
                    and ndat[self.destination_key] == dest_name
                ):
                    special_edges.append(
                        (n, ndat[self.destination_key], {self.weighing: 0})
                    )

            disrupted_graph.add_edges_from(special_edges)

            list_disrupted_destinations = []
            list_no_path = []
            node_checked_has_path = defaultdict(list)
            for n_ndat in tqdm(
                disrupted_graph.nodes.data(),
                desc=f"Finding optimal routes to {dest_name}",
            ):
                _new_origins = self._find_optimal_routes(
                    node_checked_has_path,
                    list_no_path,
                    n_ndat,
                    disrupted_graph,
                    hazard_name,
                    list_disrupted_destinations,
                    pref_routes,
                    dest_name,
                    name_save,
                    optimal_routes,
                    origins,
                    base_graph,
                    destinations,
                )

                if _new_origins is not None:
                    origins = _new_origins
                    self.get_nr_without_access(origins, list_no_path, f" {dest_name}")

            # Remove the special edges
            disrupted_graph.remove_edges_from(
                [
                    (n[0], n[1])
                    for n in disrupted_graph.edges.data()
                    if n[1] == dest_name
                ]
            )

            # Remove the closest attribute from the nodes
            for o in [
                nn[0] for nn in disrupted_graph.nodes.data() if "closest" in nn[-1]
            ]:
                del disrupted_graph.nodes[o]["closest"]

        optimal_routes_gdf = pd.concat(optimal_routes)

        return (
            base_graph,
            origins,
            destinations,
            list_disrupted_destinations,
            optimal_routes_gdf,
        )

    def load_origins(self) -> gpd.GeoDataFrame:
        od_path = self.static_path.joinpath(
            "output_graph", "origin_destination_table.feather"
        )
        od = gpd.read_feather(od_path)
        origin = od.loc[od["o_id"].notna()]
        del origin["d_id"]
        return origin

    def load_destinations(self) -> gpd.GeoDataFrame:
        od_path = self.static_path.joinpath(
            "output_graph", "origin_destination_table.feather"
        )
        od = gpd.read_feather(od_path)
        destination = od.loc[od["d_id"].notna()]
        del destination["o_id"]
        return destination

    def difference_length_with_without_hazard(
        self, with_hazard: gpd.GeoDataFrame, without_hazard: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        with_hazard.rename(columns={"length": "lengthDisr"}, inplace=True)
        without_hazard.rename(columns={"length": "lengthNorm"}, inplace=True)
        diff_length_bc_hazard = pd.merge(
            with_hazard,
            without_hazard[["lengthNorm", "origin", "destination"]],
            on=["origin", "destination"],
            how="left",
        )
        diff_length_bc_hazard["difference"] = (
            diff_length_bc_hazard["lengthDisr"].values
            - diff_length_bc_hazard["lengthNorm"].values
        )
        return diff_length_bc_hazard



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path

from ra2ce.analysis.damages.damage.damage_function_road_type_lane import (
    DamageFunctionByRoadTypeByLane,
)


class ManualDamageFunctions:
    """ "
    This class keeps an overview of the manual damage functions

    Default behaviour is to find, load and apply all available functions
    At 22 sept 2022: only implemented workflow for DamageFunction_by_RoadType_by_Lane
    """

    def __init__(self):
        self.available = (
            {}
        )  # keys = name of the available functions; values = paths to the folder
        self.loaded = []  # List of DamageFunction objects (or child classes

    def find_damage_functions(self, folder: Path) -> None:
        """Find all available damage functions in the specified folder"""
        assert folder.exists(), "Folder {} does not contain damage functions".format(
            folder
        )
        for subfolder in folder.iterdir():  # Subfolders contain the damage curves
            if subfolder.is_dir():
                self.available[subfolder.stem] = subfolder
        logging.info(
            "Found {} manual damage curves: \n {}".format(
                len(self.available.keys()), list(self.available.keys())
            )
        )
        return None

    def load_damage_functions(self):
        """ "Load damage functions in Ra2Ce"""
        for name, damage_dir in self.available.items():
            damage_function = DamageFunctionByRoadTypeByLane(name=name)
            damage_function.from_input_folder(damage_dir)
            damage_function.set_prefix()
            self.loaded.append(damage_function)
            logging.info(
                "Damage function '{}' loaded from folder {}".format(
                    damage_function.name, damage_dir
                )
            )
